__all__ = ("lookup",)

lookup = {
    "author": "This file was made by PRO100KatYT and modified for use by Delliott.",
    "Items": {
        "Schematic:ammo_bulletsheavy": {
            "name": "Heavy Bullets",
            "rarity": "common",
            "type": "ammo",
        },
        "Schematic:ammo_bulletslight": {
            "name": "Light Bullets",
            "rarity": "common",
            "type": "ammo",
        },
        "Schematic:ammo_bulletsmedium": {
            "name": "Medium Bullets",
            "rarity": "common",
            "type": "ammo",
        },
        "Schematic:ammo_energycell": {
            "name": "Energy Cells",
            "rarity": "common",
            "type": "ammo",
        },
        "Schematic:ammo_explosive": {
            "name": "Explosives",
            "rarity": "common",
            "type": "ammo",
        },
        "Schematic:ammo_shells": {
            "name": "Shells",
            "rarity": "common",
            "type": "ammo",
        },
        "Schematic:ingredient_blastpowder": {
            "name": "Blast Powder",
            "rarity": "common",
            "type": "ingredient",
        },
        "Schematic:ingredient_duct_tape": {
            "name": "Duct Tape",
            "rarity": "common",
            "type": "ingredient",
        },
        "Schematic:sid_floor_launcher_sr_t01": {
            "name": "Floor Launcher",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_tar_r_t01": {
            "name": "Tar Pit",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_ceiling_electric_single_r_t01": {
            "name": "Ceiling Zapper",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_wall_speaker_uc_t01": {
            "name": "Sound Wall",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_floor_spikes_wood_r_t01": {
            "name": "Wooden Floor Spikes",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_flamegrill_vr_t01": {
            "name": "Flame Grill Floor Trap",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_darts_uc_t01": {
            "name": "Wall Darts",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_ceiling_falling_r_t01": {
            "name": "Ceiling Drop Trap",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_wall_electric_vr_t01": {
            "name": "Wall Dynamo",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_cannons_sr_t01": {
            "name": "Broadside",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_wall_launcher_r_t01": {
            "name": "Wall Launcher",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_wall_darts_sr_t01": {
            "name": "Wall Darts",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_spikes_vr_t01": {
            "name": "Retractable Floor Spikes",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_ceiling_gas_uc_t01": {
            "name": "Ceiling Gas Trap",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_ceiling_falling_sr_t01": {
            "name": "Ceiling Drop Trap",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_launcher_r_t01": {
            "name": "Floor Launcher",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_wall_speaker_sr_t01": {
            "name": "Sound Wall",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_flamegrill_r_t01": {
            "name": "Flame Grill Floor Trap",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_campfire_r_t01": {
            "name": "Cozy Campfire",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_campfire_sr_t01": {
            "name": "Cozy Campfire",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_freeze_vr_t01": {
            "name": "Floor Freeze Trap",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_floor_ward_r_t01": {
            "name": "Anti-Air Trap",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_wall_light_r_t01": {
            "name": "Wall Lights",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_ceiling_electric_aoe_sr_t01": {
            "name": "Ceiling Electric Field",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_ward_vr_t01": {
            "name": "Anti-Air Trap",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_wood_spikes_c_t01": {
            "name": "Wall Spikes",
            "rarity": "common",
            "type": "trap",
        },
        "Schematic:sid_ceiling_electric_aoe_r_t01": {
            "name": "Ceiling Electric Field",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_wall_mechstructor_vr_t01": {
            "name": "Zap-o-max",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_electric_uc_t01": {
            "name": "Wall Dynamo",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_ceiling_gas_sr_t01": {
            "name": "Ceiling Gas Trap",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_wall_darts_r_t01": {
            "name": "Wall Darts",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_wall_wood_spikes_vr_t01": {
            "name": "Wall Spikes",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_floor_health_vr_t01": {
            "name": "Healing Pad",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_floor_freeze_r_t01": {
            "name": "Floor Freeze Trap",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_flamegrill_sr_t01": {
            "name": "Flame Grill Floor Trap",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_spikes_uc_t01": {
            "name": "Retractable Floor Spikes",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_floor_hoverboard_speed_curve_r_t01": {
            "name": "Boost Pad (Curve)",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_tar_vr_t01": {
            "name": "Tar Pit",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_electric_r_t01": {
            "name": "Wall Dynamo",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_spikes_wood_vr_t01": {
            "name": "Wooden Floor Spikes",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_light_vr_t01": {
            "name": "Wall Lights",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_ceiling_electric_single_vr_t01": {
            "name": "Ceiling Zapper",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_launcher_vr_t01": {
            "name": "Wall Launcher",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_ceiling_electric_single_c_t01": {
            "name": "Ceiling Zapper",
            "rarity": "common",
            "type": "trap",
        },
        "Schematic:sid_floor_launcher_vr_t01": {
            "name": "Floor Launcher",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_floor_spikes_wood_c_t01": {
            "name": "Wooden Floor Spikes",
            "rarity": "common",
            "type": "trap",
        },
        "Schematic:sid_wall_speaker_r_t01": {
            "name": "Sound Wall",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_wall_electric_sr_t01": {
            "name": "Wall Dynamo",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_ward_uc_t01": {
            "name": "Anti-Air Trap",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_floor_spikes_sr_t01": {
            "name": "Retractable Floor Spikes",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_wall_wood_spikes_uc_t01": {
            "name": "Wall Spikes",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_floor_health_uc_t01": {
            "name": "Healing Pad",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_floor_freeze_sr_t01": {
            "name": "Floor Freeze Trap",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_ward_sr_t01": {
            "name": "Anti-Air Trap",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_spikes_wood_uc_t01": {
            "name": "Wooden Floor Spikes",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_wall_launcher_uc_t01": {
            "name": "Wall Launcher",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_wall_cannons_vr_t01": {
            "name": "Broadside",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_darts_vr_t01": {
            "name": "Wall Darts",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_mechstructor_sr_t01": {
            "name": "Zap-o-max",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_launcher_uc_t01": {
            "name": "Floor Launcher",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_floor_defender": {
            "name": "Defender Post",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_player_jump_pad_free_direction": {
            "name": "Player Jump Pad (Directional)",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_player_jump_pad": {
            "name": "Player Jump Pad (Up)",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_hoverboard_speed_r_t01": {
            "name": "Boost Pad (Straight)",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_ceiling_gas_r_t01": {
            "name": "Ceiling Gas Trap",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_ceiling_electric_single_uc_t01": {
            "name": "Ceiling Zapper",
            "rarity": "uncommon",
            "type": "trap",
        },
        "Schematic:sid_wall_speaker_vr_t01": {
            "name": "Sound Wall",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_cannons_r_t01": {
            "name": "Broadside",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_ceiling_falling_vr_t01": {
            "name": "Ceiling Drop Trap",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_floor_health_sr_t01": {
            "name": "Healing Pad",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_health_r_t01": {
            "name": "Healing Pad",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_health_c_t00": {
            "name": "Healing Pad",
            "rarity": "common",
            "type": "trap",
        },
        "Schematic:sid_ceiling_electric_aoe_vr_t01": {
            "name": "Ceiling Electric Field",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_wood_spikes_sr_t01": {
            "name": "Wall Spikes",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_ceiling_gas_vr_t01": {
            "name": "Ceiling Gas Trap",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_floor_tar_sr_t01": {
            "name": "Tar Pit",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_spikes_r_t01": {
            "name": "Retractable Floor Spikes",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_wall_light_sr_t01": {
            "name": "Wall Lights",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_wall_wood_spikes_r_t01": {
            "name": "Wall Spikes",
            "rarity": "rare",
            "type": "trap",
        },
        "Schematic:sid_floor_campfire_vr_t01": {
            "name": "Cozy Campfire",
            "rarity": "epic",
            "type": "trap",
        },
        "Schematic:sid_wall_launcher_sr_t01": {
            "name": "Wall Launcher",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_floor_spikes_wood_sr_t01": {
            "name": "Wooden Floor Spikes",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_ceiling_electric_single_sr_t01": {
            "name": "Ceiling Zapper",
            "rarity": "legendary",
            "type": "trap",
        },
        "Schematic:sid_assault_auto_halloween_sr_ore_t01": {
            "name": "Grave Digger",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_dragon_vr_ore_t01": {
            "name": "Dragon's Fang",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_autodrum_sr_ore_t01": {
            "name": "Typewriter",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_fleming_vr_ore_t01": {
            "name": "The Browbeater",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_hydraulic_vr_ore_t01": {
            "name": "Piston Spitter",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_blackmetal_r_ore_t01": {
            "name": "Nightfall",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_sniper_shredder_vr_ore_t01": {
            "name": "Super Shredder",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_scavenger_sr_ore_t01": {
            "name": "Pipe Down!",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_boombox_drum_sr_ore_t01": {
            "name": "Staccato Shotto",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_ratrod_vr_ore_t01": {
            "name": "Crankshot",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_uc_ore_t01": {
            "name": "Double-punch",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_medieval_sr_ore_t01": {
            "name": "Stalwart Squire",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_lmg_drum_founders_sr_ore_t01": {
            "name": "Founder's Drumroll",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_auto_c_ore_t01": {
            "name": "Viper",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_pistol_autoheavy_vt_sr_ore_t01": {
            "name": "Vindertech Blazer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medium_c_ore_t01": {
            "name": "Handmade Axe",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_explosive_invasion_plasmacannon_sr_ore_t01": {
            "name": "Plasmatic Discharger",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_snowball_vr_ore_t01": {
            "name": "Blizzard Blitzer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_neonglow_vr_ore_t01": {
            "name": "Krypton Pistol",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_scope_vt_sr_ore_t01": {
            "name": "Vindertech Jolter",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_retroscifi_sr_ore_t01": {
            "name": "Plasmatron 9000",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_heavy_vr_ore_t01": {
            "name": "Stormchaser's Revenge",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_flintlock_sr_ore_t01": {
            "name": "Dragoon",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_dragon_sr_ore_t01": {
            "name": "Dragon's Breath",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_military_vr_ore_t01": {
            "name": "Ground Pounder",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_primal_stinkbow_sr_ore_t01": {
            "name": "Primal Stink Bow",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_icesword_sr_ore_t01": {
            "name": "Slicicle",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_de_sr_ore_t01": {
            "name": "Bald Eagle",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_revolver_singleaction_sr_ore_t01": {
            "name": "Last Word",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_shovel_halloween_vr_ore_t01": {
            "name": "Six Feet Under",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_heavy_r_ore_t01": {
            "name": "Greataxe",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_explosive_steampunk_sr_ore_t01": {
            "name": "Cannonade",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_boombox_break_sr_ore_t01": {
            "name": "Woofer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_burst_retroscifi_sr_ore_t01": {
            "name": "Gammatron 9000",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_r_ore_t01": {
            "name": "Judge",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_rocket_vt_sr_ore_t01": {
            "name": "Vindertech Slammer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_laser_vr_ore_t01": {
            "name": "Reaper Scythe",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_military_auto_sr_ore_t01": {
            "name": "Bobcat",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_retroscifi_sr_ore_t01": {
            "name": "De-Atomizer 9000",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_boombox_vr_ore_t01": {
            "name": "Crescendo",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_scavenger_sr_ore_t01": {
            "name": "Sunsetter",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_laser_founders_vr_ore_t01": {
            "name": "Founder's Blazing Masamune",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_neonglow_vr_ore_t01": {
            "name": "Argon Assault Rifle",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_zapper_vr_ore_t01": {
            "name": "Zap Zapp",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_hydraulic_sr_ore_t01": {
            "name": "Dam Buster",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_blackmetal_bow_sr_ore_t01": {
            "name": "Night Owl",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_heavy_vacuumtube_vr_ore_t01": {
            "name": "Vacuum Tube Axe",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_steampunk_vr_ore_t01": {
            "name": "Steamroller",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_burst_r_ore_t01": {
            "name": "Razor",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_boombox_vr_ore_t01": {
            "name": "Earsplitter",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_rocketbat_vr_ore_t01": {
            "name": "Fortsville Slugger 3000",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_artdeco_sr_ore_t01": {
            "name": "The Floor Flusher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_pirate_sr_ore_t01": {
            "name": "Corsair",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_scavenger_sr_ore_t01": {
            "name": "Head Clipper",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_heavy_vr_ore_t01": {
            "name": "Doomhammer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_retroscifi_sr_ore_t01": {
            "name": "Atomic Light Expander",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_hydraulic_drum_vr_ore_t01": {
            "name": "Nailer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_medium_r_ore_t01": {
            "name": "Heavy Wrench",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_steampunk_sr_ore_t01": {
            "name": "Steam Thrasher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_bat_r_ore_t01": {
            "name": "Mighty Slugger",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_launcher_rocket_vr_ore_t01": {
            "name": "Bazooka",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_raygun_vr_ore_t01": {
            "name": "Deathray",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_uc_ore_t01": {
            "name": "Shadow Blade",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_sniper_valentine_bow_sr_ore_t01": {
            "name": "Love Song",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_rocket_vr_ore_t01": {
            "name": "Husk Stomper",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_military_silenced_vr_ore_t01": {
            "name": "Wraith",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_flintlock_sr_ore_t01": {
            "name": "Walloper",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_autoheavy_sr_ore_t01": {
            "name": "Thrasher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_r_ore_t01": {
            "name": "Mighty Driver",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_vt_vr_ore_t01": {
            "name": "Vindertech Sever",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_boltaction_scope_vr_ore_t01": {
            "name": "One Shot",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_sixshooter_uc_ore_t01": {
            "name": "Nighthawk",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_christmas_sr_ore_t01": {
            "name": "Tree of Light",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_winter_sr_ore_t01": {
            "name": "Frostbite",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_military_sr_ore_t01": {
            "name": "Ground Pounder",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_heavy_c_ore_t01": {
            "name": "Stop Axe",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_sniper_dragon_vr_ore_t01": {
            "name": "Dragon's Claw",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_dragon_vr_ore_t01": {
            "name": "Dragon's Might",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_autodrum_teddy_vr_ore_t01": {
            "name": "Papa Bear",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_ratrod_vr_ore_t01": {
            "name": "Two Step",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_hydraulic_sr_ore_t01": {
            "name": "Pummeller",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_shovel_halloween_sr_ore_t01": {
            "name": "Six Feet Under",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_dragon_sr_ore_t01": {
            "name": "Dragon's Fang",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_heavy_sr_ore_t01": {
            "name": "Stormchaser's Revenge",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_heavy_founders_vr_ore_t01": {
            "name": "Founder's Warhammer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_blackmetal_sr_ore_t01": {
            "name": "Nightfall",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_artdeco_vr_ore_t01": {
            "name": "The Killjoy",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_tactical_founders_r_ore_t01": {
            "name": "Founder's Deconstructor",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_boombox_vr_ore_t01": {
            "name": "Beat Blaster",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_ratrod_sr_ore_t01": {
            "name": "Crankshot",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_auto_vr_ore_t01": {
            "name": "Siegebreaker",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_boombox_r_ore_t01": {
            "name": "Mic Drop",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_snowball_sr_ore_t01": {
            "name": "Blizzard Blitzer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_vacuumtube_vr_ore_t01": {
            "name": "Vacuum Tube Spear",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_blackmetal_sr_ore_t01": {
            "name": "Midnight Stalker",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_military_burst_sr_ore_t01": {
            "name": "Lynx",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_neonglow_sr_ore_t01": {
            "name": "Krypton Pistol",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_auto_r_ore_t01": {
            "name": "Extinguisher",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_dragon_vr_ore_t01": {
            "name": "Dragon's Tooth",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_steampunk_vr_ore_t01": {
            "name": "The Baron",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_artdeco_sr_ore_t01": {
            "name": "The Mouth Piece",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_founders_vr_ore_t01": {
            "name": "Founder's Wasp",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_vr_ore_t01": {
            "name": "Dragonfly",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_artdeco_vr_ore_t01": {
            "name": "The Brush Off",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_steampunk_sr_ore_t01": {
            "name": "Steam Skewer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_rocketbat_sr_ore_t01": {
            "name": "Fortsville Slugger 3000",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_neonglow_sr_ore_t01": {
            "name": "Argon Assault Rifle",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_auto_founders_sr_ore_t01": {
            "name": "Nocturno",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_hydra_sr_ore_t01": {
            "name": "Hydra",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_hydraulic_sr_ore_t01": {
            "name": "Piston Spitter",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_rapid_founders_vr_ore_t01": {
            "name": "Founder's Revolt",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_rocket_sr_ore_t01": {
            "name": "Bazooka",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_scavenger_vr_ore_t01": {
            "name": "Old Smokey",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_military_vr_ore_t01": {
            "name": "Dissector",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_boxing_glove_club_vr_ore_t01": {
            "name": "The Contender",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_bbgun_r_ore_t01": {
            "name": "Ralphie's Revenge",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_standard_c_ore_t01": {
            "name": "Knight",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_pistol_paper_airplane_sr_ore_t01": {
            "name": "The Paper Shredder",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_vacuumtube_bow_sr_ore_t01": {
            "name": "Vacuum Tube Bow",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_artdeco_break_vr_ore_t01": {
            "name": "The Big Shot",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_neonglow_vr_ore_t01": {
            "name": "Neon Scythe",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medieval_sr_ore_t01": {
            "name": "Guardian's Will",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_dragon_r_ore_t01": {
            "name": "Dragon's Claw",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_explosive_launcher_military_vr_ore_t01": {
            "name": "Quad Launcher",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_vr_ore_t01": {
            "name": "Wrecking Bar",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_shredder_sr_ore_t01": {
            "name": "Super Shredder",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_raygun_sr_ore_t01": {
            "name": "Deathray",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_lmg_vr_ore_t01": {
            "name": "Terminator",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_amr_vr_ore_t01": {
            "name": "Zapotron",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_auto_vr_ore_t01": {
            "name": "Deathwing",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_laser_founders_sr_ore_t01": {
            "name": "Founder's Blazing Masamune",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_firecracker_vr_ore_t01": {
            "name": "Freedom's Herald",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_dragon_drum_vr_ore_t01": {
            "name": "DragonFire",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_standard_vt_vr_ore_t01": {
            "name": "Vindertech Disintegrator",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_zapper_sr_ore_t01": {
            "name": "Zap Zapp",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_heavy_sr_ore_t01": {
            "name": "Doomhammer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_light_r_ore_t01": {
            "name": "Shortblade Slicer",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_smg_fleming_burstfire_vr_ore_t01": {
            "name": "The Trinity",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_doubleshot_vr_ore_t01": {
            "name": "Ranger",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_laser_sr_ore_t01": {
            "name": "Reaper Scythe",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_vacuumtube_auto_vr_ore_t01": {
            "name": "Lightning Pistol",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_heavy_vacuumtube_sr_ore_t01": {
            "name": "Vacuum Tube Axe",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_boombox_sr_ore_t01": {
            "name": "Earsplitter",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_flintlock_bow_sr_ore_t01": {
            "name": "The Powder Keg",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_standard_vr_ore_t01": {
            "name": "Templar",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_steampunk_sr_ore_t01": {
            "name": "Steamroller",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_fleming_sr_ore_t01": {
            "name": "The Browbeater",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_ou_vr_ore_t01": {
            "name": "Maverick",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_flintlock_scope_vr_ore_t01": {
            "name": "Spyglass",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_r_ore_t01": {
            "name": "Double-punch",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_sniper_military_vr_ore_t01": {
            "name": "Obliterator",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_boltaction_scope_sr_ore_t01": {
            "name": "One Shot",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_auto_sr_ore_t01": {
            "name": "Siegebreaker",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_semiauto_vr_ore_t01": {
            "name": "Jackal",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_ratrod_sr_ore_t01": {
            "name": "Two Step",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_heavy_vr_ore_t01": {
            "name": "Huskcleaver",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_medium_vr_ore_t01": {
            "name": "Socket Slugger",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_longarm_vr_ore_t01": {
            "name": "Longarm Enforcer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_laser_vr_ore_t01": {
            "name": "Stormblade",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_lmg_military_vr_ore_t01": {
            "name": "HackSAW",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_dragon_sr_ore_t01": {
            "name": "Dragon's Tooth",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_launcher_pumpkin_rpg_sr_ore_t01": {
            "name": "Jack-O-Launcher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_light_ratrod_vr_ore_t01": {
            "name": "Axle",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_retroscifi_vr_ore_t01": {
            "name": "Astro Bat 9000",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_autodrum_teddy_sr_ore_t01": {
            "name": "Papa Bear",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_vacuumtube_revolver_vr_ore_t01": {
            "name": "Vacuum Tube Revolver",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_dragon_sr_ore_t01": {
            "name": "Dragon's Might",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_vacuumtube_sr_ore_t01": {
            "name": "Vacuum Tube Spear",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_boxing_glove_club_r_ore_t01": {
            "name": "The Contender",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_hydraulic_drum_sr_ore_t01": {
            "name": "Nailer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_military_silenced_sr_ore_t01": {
            "name": "Wraith",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_artdeco_vr_ore_t01": {
            "name": "The Great Axeby",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_heavy_founders_vr_ore_t01": {
            "name": "Founder's Greatsword",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_boombox_sr_ore_t01": {
            "name": "Beat Blaster",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_bolt_vr_ore_t01": {
            "name": "Bolt Bolt",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_auto_uc_ore_t01": {
            "name": "Viper",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_artdeco_sr_ore_t01": {
            "name": "The Killjoy",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_scope_vr_ore_t01": {
            "name": "Scoped Dragonfly",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_boombox_sr_ore_t01": {
            "name": "Crescendo",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_medieval_smg_vr_ore_t01": {
            "name": "Hemlock",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_vr_ore_t01": {
            "name": "Elegant Scythe",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_steampunk_sr_ore_t01": {
            "name": "The Baron",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_scavenger_sr_ore_t01": {
            "name": "Old Smokey",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_vt_sr_ore_t01": {
            "name": "Vindertech Sever",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_dragon_r_ore_t01": {
            "name": "Dragon's Roar",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_auto_vr_ore_t01": {
            "name": "Viper",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_vacuumtube_vr_ore_t01": {
            "name": "Vacuum Tube Launcher",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_airplane_axe_vr_ore_t01": {
            "name": "Supersonic Slasher",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_artdeco_sr_ore_t01": {
            "name": "The Brush Off",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medium_uc_ore_t01": {
            "name": "Splinter Axe",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_sniper_steampunk_vr_ore_t01": {
            "name": "Bon Voyage",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_standard_sr_ore_t01": {
            "name": "The Bear",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_singleburst_blackmetal_sr_ore_t01": {
            "name": "Dirge Song",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_smg_silenced_military_vr_ore_t01": {
            "name": "Silenced Specter",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_burst_blackmetal_r_ore_t01": {
            "name": "Black Drum",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_sniper_scavenger_bow_sr_ore_t01": {
            "name": "Scrap Shot",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_dragon_sr_ore_t01": {
            "name": "Dragon's Claw",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_artdeco_vr_ore_t01": {
            "name": "The Heavyweight",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_rocket_sr_ore_t01": {
            "name": "Husk Stomper",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_heavy_r_ore_t01": {
            "name": "Heavy Sledge",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_doubleshot_sr_ore_t01": {
            "name": "Ranger",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_heavy_uc_ore_t01": {
            "name": "Fractured Axe",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_sr_ore_t01": {
            "name": "Dragonfly",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_burst_c_ore_t01": {
            "name": "Razor",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_pistol_flintlock_vr_ore_t01": {
            "name": "Duelist",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_dragon_drum_sr_ore_t01": {
            "name": "DragonFire",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_firecracker_sr_ore_t01": {
            "name": "Freedom's Herald",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_neonglow_sr_ore_t01": {
            "name": "Neon Scythe",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_standard_vt_sr_ore_t01": {
            "name": "Vindertech Disintegrator",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_auto_sr_ore_t01": {
            "name": "Deathwing",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_boxing_glove_club_sr_ore_t01": {
            "name": "The Contender",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_ratrod_vr_ore_t01": {
            "name": "Lead Sled",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_explosive_launcher_military_sr_ore_t01": {
            "name": "Quad Launcher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_auto_r_ore_t01": {
            "name": "Tigerjaw",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_semiauto_founders_vr_ore_t01": {
            "name": "Founder's Raptor",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_rapid_vr_ore_t01": {
            "name": "Haywire Storm",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_sr_ore_t01": {
            "name": "Mr. Red",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_hydraulic_vr_ore_t01": {
            "name": "Tsunami",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_boltrevolver_uc_ore_t01": {
            "name": "Junkyard Pal",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_pistol_vacuumtube_auto_sr_ore_t01": {
            "name": "Lightning Pistol",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_artdeco_sr_ore_t01": {
            "name": "The Great Axeby",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_pirate_sr_ore_t01": {
            "name": "Jack's Revenge",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_light_ratrod_sr_ore_t01": {
            "name": "Axle",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_amr_sr_ore_t01": {
            "name": "Zapotron",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_boltrevolver_r_ore_t01": {
            "name": "Junkyard Pal",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_lmg_sr_ore_t01": {
            "name": "Terminator",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_semiauto_vr_ore_t01": {
            "name": "Hunter-killer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_r_ore_t01": {
            "name": "Reaver",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_blunt_medium_sr_ore_t01": {
            "name": "Socket Slugger",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_singleshot_r_ore_t01": {
            "name": "Hammercrush",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_vr_ore_t01": {
            "name": "Judge",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medium_r_ore_t01": {
            "name": "Splinter Axe",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_sniper_bbgun_vr_ore_t01": {
            "name": "Ralphie's Revenge",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_ou_uc_ore_t01": {
            "name": "Rustler",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_semiauto_scavenger_vr_ore_t01": {
            "name": "Husk Buster",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_piercing_junkyard_vr_ore_t01": {
            "name": "Scrapper Spear",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_uc_ore_t01": {
            "name": "Mighty Pitchfork",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_halloween_sr_ore_t01": {
            "name": "Spectral Blade",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_semiauto_r_ore_t01": {
            "name": "Vigilante",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_smg_fleming_burstfire_sr_ore_t01": {
            "name": "The Trinity",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_surgical_vr_ore_t01": {
            "name": "Deathstalker",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_autoheavy_r_ore_t01": {
            "name": "Thrasher",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_light_uc_ore_t01": {
            "name": "Shortblade Slicer",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_tactical_r_ore_t01": {
            "name": "Panther",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_medieval_smg_sr_ore_t01": {
            "name": "Hemlock",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_vacuumtube_revolver_sr_ore_t01": {
            "name": "Vacuum Tube Revolver",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_lmg_r_ore_t01": {
            "name": "Terminator",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_semiauto_sr_ore_t01": {
            "name": "Jackal",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_sr_ore_t01": {
            "name": "Elegant Scythe",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_tripleshot_vr_ore_t01": {
            "name": "Triple Tap",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_lmg_military_sr_ore_t01": {
            "name": "HackSAW",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_laser_vr_ore_t01": {
            "name": "Farmer's Glory",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_artdeco_break_sr_ore_t01": {
            "name": "The Big Shot",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_heavy_sr_ore_t01": {
            "name": "Huskcleaver",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_semiauto_uc_ore_t01": {
            "name": "Stampede",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_military_sr_ore_t01": {
            "name": "Dissector",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_ou_r_ore_t01": {
            "name": "Rustler",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_ou_sr_ore_t01": {
            "name": "Maverick",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_neonglow_vr_ore_t01": {
            "name": "Krypton Sword",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_halloween_handcannon_sr_ore_t01": {
            "name": "Ghost Pistol",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_light_c_ore_t01": {
            "name": "Shortblade",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medium_vr_ore_t01": {
            "name": "Revenant's Wrath",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_scope_sr_ore_t01": {
            "name": "Scoped Dragonfly",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_heavy_dragon_vr_ore_t01": {
            "name": "Dragon's Fist",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_auto_r_ore_t01": {
            "name": "Breacher",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_auto_sr_ore_t01": {
            "name": "Viper",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_c_ore_t01": {
            "name": "Simple Pitchfork",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_pistol_flintlock_sr_ore_t01": {
            "name": "Duelist",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_rapid_sr_ore_t01": {
            "name": "Haywire Storm",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_vr_ore_t01": {
            "name": "Whirling Doom",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_smg_fleming_sidefeed_vr_ore_t01": {
            "name": "The Pepper Sprayer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_semiauto_steampunk_vr_ore_t01": {
            "name": "Steam Piston",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medium_laser_vr_ore_t01": {
            "name": "Armaggeddon",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_rocket_r_ore_t01": {
            "name": "Bazooka",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_explosive_blackmetal_sr_ore_t01": {
            "name": "Metal Marauder",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_boltaction_r_ore_t01": {
            "name": "Deadeye",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_retroscifi_sr_ore_t01": {
            "name": "Astro Bat 9000",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_c_ore_t01": {
            "name": "Hockey Stick",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_sniper_retroscifi_vr_ore_t01": {
            "name": "Blastotron 9000",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_vacuumtube_sr_ore_t01": {
            "name": "Vacuum Tube Launcher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_ratrod_sr_ore_t01": {
            "name": "Lead Sled",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_semi_vt_vr_ore_t01": {
            "name": "Vindertech Burster",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_heavy_flintlock_vr_ore_t01": {
            "name": "Sir Hootie",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_longarm_sr_ore_t01": {
            "name": "Longarm Enforcer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_bolt_sr_ore_t01": {
            "name": "Bolt Bolt",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_launcher_rocket_dragon_vr_ore_t01": {
            "name": "Dragon's Fury",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_military_sr_ore_t01": {
            "name": "Obliterator",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_laser_sr_ore_t01": {
            "name": "Stormblade",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_vacuumtube_vr_ore_t01": {
            "name": "Vacuum Tube Sniper Rifle",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_medium_uc_ore_t01": {
            "name": "Heavy Wrench",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_assault_shockwave_vr_ore_t01": {
            "name": "Lectro Shock Rifle",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_blackmetal_sr_ore_t01": {
            "name": "Brood Slicer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_ratrod_slug_sr_ore_t01": {
            "name": "Rat King",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_standard_uc_ore_t01": {
            "name": "Knight",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_edged_airplane_axe_sr_ore_t01": {
            "name": "Supersonic Slasher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_flintlock_scope_sr_ore_t01": {
            "name": "Spyglass",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_steampunk_vr_ore_t01": {
            "name": "Double Boiler",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_tripleshot_sr_ore_t01": {
            "name": "Triple Tap",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_boltaction_scope_r_ore_t01": {
            "name": "One Shot",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_laser_sr_ore_t01": {
            "name": "Farmer's Glory",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_launcher_grenade_easter_sr_ore_t01": {
            "name": "Easter Egg Launcher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_semiauto_scavenger_sr_ore_t01": {
            "name": "Husk Buster",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_boombox_sr_ore_t01": {
            "name": "Deciblaster",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_tactical_uc_ore_t01": {
            "name": "Panther",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_semi_vr_ore_t01": {
            "name": "Falcon",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_auto_vr_ore_t01": {
            "name": "Tigerjaw",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_fleming_vr_ore_t01": {
            "name": "The Cleaner",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_c_ore_t01": {
            "name": "Double-punch",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_c_ore_t01": {
            "name": "Baseball Bat",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_sr_ore_t01": {
            "name": "Judge",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_medieval_bow_sr_ore_t01": {
            "name": "Knightfire",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_medieval_vr_ore_t01": {
            "name": "Basilisk",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_artdeco_sr_ore_t01": {
            "name": "The Heavyweight",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_retroscifi_sr_ore_t01": {
            "name": "Pulsar 9000",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_hydraulic_sr_ore_t01": {
            "name": "Tsunami",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_vacuumtube_vr_ore_t01": {
            "name": "Vacuum Tube Sword",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_semiauto_sr_ore_t01": {
            "name": "Hunter-killer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_crossbow_sr_ore_t01": {
            "name": "Heartbreaker",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_heavy_sr_ore_t01": {
            "name": "Backbreaker",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_steampunk_sr_ore_t01": {
            "name": "Bon Voyage",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_semiauto_uc_ore_t01": {
            "name": "Vigilante",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_heavy_r_ore_t01": {
            "name": "Highlander",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medieval_vr_ore_t01": {
            "name": "Wicked Wind",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_uc_ore_t01": {
            "name": "Stinger",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_light_founders_vr_ore_t01": {
            "name": "Founder's Nightcleave",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_hydraulic_bow_sr_ore_t01": {
            "name": "Compression Burster",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_semiauto_uc_ore_t01": {
            "name": "Predator",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_assault_smg_silenced_military_sr_ore_t01": {
            "name": "Silenced Specter",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_boombox_vr_ore_t01": {
            "name": "Mic Drop",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_flintlock_vr_ore_t01": {
            "name": "Ol' Betsy",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_winter_sr_ore_t01": {
            "name": "Pain Train",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_medieval_vr_ore_t01": {
            "name": "Smasher Basher",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_hydraulic_vr_ore_t01": {
            "name": "Sawtooth",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_medieval_sr_ore_t01": {
            "name": "Sir Lancelot",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_silenced_military_vr_ore_t01": {
            "name": "Whisper .45",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_artdeco_vr_ore_t01": {
            "name": "The Take Down",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_club_light_vr_ore_t01": {
            "name": "Masters Driver",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_r_ore_t01": {
            "name": "Stinger",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_rocket_sr_ore_t01": {
            "name": "Tiny Instrument of Death",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_scavenger_vr_ore_t01": {
            "name": "Hack Axe",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_neonglow_vr_ore_t01": {
            "name": "Helium Shotgun",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medium_sr_ore_t01": {
            "name": "Revenant's Wrath",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_military_r_ore_t01": {
            "name": "Dissector",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_semi_vt_sr_ore_t01": {
            "name": "Vindertech Burster",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_auto_uc_ore_t01": {
            "name": "Breacher",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_smg_fleming_sidefeed_sr_ore_t01": {
            "name": "The Pepper Sprayer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_tactical_precision_r_ore_t01": {
            "name": "Nightclaw",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_sniper_vacuumtube_sr_ore_t01": {
            "name": "Vacuum Tube Sniper Rifle",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_neonglow_sr_ore_t01": {
            "name": "Krypton Sword",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_autoheavy_founders_vr_ore_t01": {
            "name": "Founder's QuickShot",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_surgical_sr_ore_t01": {
            "name": "Deathstalker",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medium_laser_sr_ore_t01": {
            "name": "Armaggeddon",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_junkyard_sr_ore_t01": {
            "name": "Scrapper Spear",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_burst_vr_ore_t01": {
            "name": "Razorblade",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_flintlock_r_ore_t01": {
            "name": "Duelist",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_blunt_heavy_flintlock_sr_ore_t01": {
            "name": "Sir Hootie",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_smg_invasion_raygun_sr_ore_t01": {
            "name": "Shooting Star",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_bbgun_sr_ore_t01": {
            "name": "Ralphie's Revenge",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_heavy_paddle_r_ore_t01": {
            "name": "The Initiator",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_stormking_sr_ore_t01": {
            "name": "Storm King's Onslaught",
            "rarity": "mythic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_rocket_dragon_sr_ore_t01": {
            "name": "Dragon's Fury",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_neonglow_vr_ore_t01": {
            "name": "Neon Sniper Rifle",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_c_ore_t01": {
            "name": "Ruler Sword",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_tactical_founders_vr_ore_t01": {
            "name": "Founder's Deconstructor",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_auto_retroscifi_vr_ore_t01": {
            "name": "Blastotron Mini",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_stormking_sr_ore_t01": {
            "name": "Storm King's Ravager",
            "rarity": "mythic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_uc_ore_t01": {
            "name": "Mighty Driver",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_sniper_neon_bow_sr_ore_t01": {
            "name": "Xenon Bow",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_neonglow_vr_ore_t01": {
            "name": "Argon Axe",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_retroscifi_sr_ore_t01": {
            "name": "Blastotron 9000",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_scavenger_vr_ore_t01": {
            "name": "Riptide",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_semiauto_c_ore_t01": {
            "name": "Vigilante",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medieval_sr_ore_t01": {
            "name": "Wicked Wind",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_launcher_ratrod_sr_ore_t01": {
            "name": "V6 Launcher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_tactical_c_ore_t01": {
            "name": "Panther",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_heavy_dragon_sr_ore_t01": {
            "name": "Dragon's Fist",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_auto_sr_ore_t01": {
            "name": "Tigerjaw",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_medium_ratrod_vr_ore_t01": {
            "name": "Drifter",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_blackmetal_sr_ore_t01": {
            "name": "Black Blade",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_boltrevolver_c_ore_t01": {
            "name": "Junkyard Pal",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_sniper_auto_founders_vr_ore_t01": {
            "name": "Founder's Exterminator",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_gatling_vr_ore_t01": {
            "name": "Monsoon",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_boombox_sr_ore_t01": {
            "name": "Mic Drop",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_vr_ore_t01": {
            "name": "Slice n' Dice",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_steampunk_sr_ore_t01": {
            "name": "Double Boiler",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_light_r_ore_t01": {
            "name": "Ski Cleaver",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_semiauto_r_ore_t01": {
            "name": "Predator",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_sniper_fleming_sr_ore_t01": {
            "name": "The Cleaner",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_launcher_grenade_winter_sr_ore_t01": {
            "name": "Snowball Launcher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_semiauto_steampunk_sr_ore_t01": {
            "name": "Steam Piston",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_sr_ore_t01": {
            "name": "Whirling Doom",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_boombox_sr_ore_t01": {
            "name": "Hot Mix",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_artdeco_vr_ore_t01": {
            "name": "The Sod Buster",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_shockwave_sr_ore_t01": {
            "name": "Lectro Shock Rifle",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_heavy_uc_ore_t01": {
            "name": "Sledgehammer",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_r_ore_t01": {
            "name": "Shadow Blade",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_medieval_sr_ore_t01": {
            "name": "Basilisk",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_boltaction_uc_ore_t01": {
            "name": "Deadeye",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_blunt_medium_c_ore_t01": {
            "name": "Trusty Pipe",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_vacuumtube_sr_ore_t01": {
            "name": "Vacuum Tube Sword",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_sixshooter_r_ore_t01": {
            "name": "Nighthawk",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_dragon_r_ore_t01": {
            "name": "Dragon's Tooth",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_blunt_tool_light_uc_ore_t01": {
            "name": "Breaching Bar",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_uc_ore_t01": {
            "name": "Reaver",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_scavenger_sr_ore_t01": {
            "name": "Hack Axe",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_neonglow_sr_ore_t01": {
            "name": "Helium Shotgun",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_rapid_r_ore_t01": {
            "name": "Haywire Storm",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_sniper_boltaction_c_ore_t01": {
            "name": "Deadeye",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_stormking_sr_ore_t01": {
            "name": "Storm King's Fury",
            "rarity": "mythic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_minigun_sr_ore_t01": {
            "name": "Room Sweeper",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_burst_blackmetal_vr_ore_t01": {
            "name": "Black Drum",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_military_surgical_vr_ore_t01": {
            "name": "Tiger",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_medieval_sr_ore_t01": {
            "name": "Smasher Basher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_steampunk_drum_vr_ore_t01": {
            "name": "Duet",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_vindertech_bow_sr_ore_t01": {
            "name": "Vindertech Seeker",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_vt_sr_ore_t01": {
            "name": "Vindertech Blaster",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_neonglow_lmg_vr_ore_t01": {
            "name": "Mercury LMG",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_grenade_vr_ore_t01": {
            "name": "Thumper",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_retroscifi_vr_ore_t01": {
            "name": "Husk Warper",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_flintlock_vr_ore_t01": {
            "name": "Fishing Hook",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_singleshot_vr_ore_t01": {
            "name": "Hammercrush",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_ratrod_autodrum_vr_ore_t01": {
            "name": "Swan",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_burst_uc_ore_t01": {
            "name": "Razor",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_semi_sr_ore_t01": {
            "name": "Falcon",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_winter_r_ore_t01": {
            "name": "Popshot",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_space_vr_ore_t01": {
            "name": "Space Invader",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_vt_vr_ore_t01": {
            "name": "Vindertech Slicer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_scavenger_vr_ore_t01": {
            "name": "Trash Cannon",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_auto_c_ore_t01": {
            "name": "Breacher",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_sniper_artdeco_sr_ore_t01": {
            "name": "The Take Down",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_launcher_neonglow_vr_ore_t01": {
            "name": "Noble Launcher",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_auto_vt_vr_ore_t01": {
            "name": "Vindertech Pulsar",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_flintlock_vr_ore_t01": {
            "name": "Bundlebuss",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_singleshot_blackmetal_vr_ore_t01": {
            "name": "Blackout",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_founders_vr_ore_t01": {
            "name": "Founder's Lawbringer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_tactical_precision_vr_ore_t01": {
            "name": "Nightclaw",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_crossbow_military_vr_ore_t01": {
            "name": "Crossbow",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_explosive_stormking_sr_ore_t01": {
            "name": "Storm King's Wrath",
            "rarity": "mythic",
            "type": "weapon",
        },
        "Schematic:sid_assault_boombox_r_ore_t01": {
            "name": "Beat Blaster",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_silenced_military_sr_ore_t01": {
            "name": "Whisper .45",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_vr_ore_t01": {
            "name": "Thunderbolt",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_semiauto_vr_ore_t01": {
            "name": "Stampede",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_autoheavy_founders_r_ore_t01": {
            "name": "Founder's QuickShot",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_launcher_flintlock_vr_ore_t01": {
            "name": "Bowler",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_grenade_r_ore_t01": {
            "name": "Thumper",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_lmg_halloween_sr_ore_t01": {
            "name": "Candy Corn LMG",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_flintlock_sr_ore_t01": {
            "name": "Ol' Betsy",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_winter_gingerbread_vr_ore_t01": {
            "name": "Ginger Blaster",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_neonglow_sr_ore_t01": {
            "name": "Argon Axe",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_fleming_scoped_vr_ore_t01": {
            "name": "The Cyclops",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_burst_sr_ore_t01": {
            "name": "Razorblade",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_shark_sr_ore_t01": {
            "name": "Shark Attack!",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_light_vr_ore_t01": {
            "name": "Stabsworth the III",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_autoheavy_founders_sr_ore_t01": {
            "name": "Founder's QuickShot",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_tactical_founders_sr_ore_t01": {
            "name": "Founder's Deconstructor",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_hydraulic_vr_ore_t01": {
            "name": "Pulverizer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_blackmetal_vr_ore_t01": {
            "name": "Brood Slicer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_club_light_sr_ore_t01": {
            "name": "Masters Driver",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_r_ore_t01": {
            "name": "Mighty Pitchfork",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_auto_r_ore_t01": {
            "name": "Viper",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_boombox_break_r_ore_t01": {
            "name": "Woofer",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_hydraulic_sr_ore_t01": {
            "name": "Sawtooth",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_winter_vr_ore_t01": {
            "name": "Popshot",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_auto_retroscifi_sr_ore_t01": {
            "name": "Blastotron Mini",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_heavy_paddle_uc_ore_t01": {
            "name": "The Initiator",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_auto_uc_ore_t01": {
            "name": "Tigerjaw",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_pistol_fleming_tactical_vr_ore_t01": {
            "name": "The Turncoat",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_grenade_sr_ore_t01": {
            "name": "Thumper",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_boombox_break_vr_ore_t01": {
            "name": "Woofer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_ratrod_sr_ore_t01": {
            "name": "Ratatat",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_vacuumtube_vr_ore_t01": {
            "name": "Vacuum Tube Shotgun",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_burst_retroscifi_vr_ore_t01": {
            "name": "Gammatron 9000",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_space_sr_ore_t01": {
            "name": "Space Invader",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_burst_blackmetal_sr_ore_t01": {
            "name": "Black Drum",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_dragon_vr_ore_t01": {
            "name": "Dragon's Breath",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_c_ore_t01": {
            "name": "Stinger",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_vt_sr_ore_t01": {
            "name": "Vindertech Slicer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_fleming_vr_ore_t01": {
            "name": "The Grunt",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_hydraulic_vr_ore_t01": {
            "name": "Pressure Cutter",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_stormking_sr_ore_t01": {
            "name": "Storm King's Scourge",
            "rarity": "mythic",
            "type": "weapon",
        },
        "Schematic:sid_assault_vacuumtube_vr_ore_t01": {
            "name": "Vacuum Tube Rifle",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_heavy_paddle_c_ore_t01": {
            "name": "The Initiator",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_semi_r_ore_t01": {
            "name": "Falcon",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_retroscifi_vr_ore_t01": {
            "name": "Plasmatron 9000",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_light_vr_ore_t01": {
            "name": "Claxe",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_flintlock_sr_ore_t01": {
            "name": "Bundlebuss",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_standard_scope_vt_vr_ore_t01": {
            "name": "Vindertech Jolter",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_gatling_sr_ore_t01": {
            "name": "Monsoon",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_scavenger_vr_ore_t01": {
            "name": "Buzzcut",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_dragon_vr_ore_t01": {
            "name": "Dragon's Roar",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_boombox_drum_vr_ore_t01": {
            "name": "Staccato Shotto",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_lmg_drum_founders_vr_ore_t01": {
            "name": "Founder's Drumroll",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_sr_ore_t01": {
            "name": "Slice n' Dice",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_medium_ratrod_sr_ore_t01": {
            "name": "Drifter",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_scavenger_sr_ore_t01": {
            "name": "Riptide",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_military_surgical_sr_ore_t01": {
            "name": "Tiger",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_artdeco_sr_ore_t01": {
            "name": "The Sod Buster",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_retroscifi_sr_ore_t01": {
            "name": "Husk Warper",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_fleming_sr_ore_t01": {
            "name": "The Potshot",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_flintlock_r_ore_t01": {
            "name": "Ol' Betsy",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_singleshot_sr_ore_t01": {
            "name": "Hammercrush",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_neonglow_lmg_sr_ore_t01": {
            "name": "Mercury LMG",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_heavy_uc_ore_t01": {
            "name": "Highlander",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_sniper_fleming_bow_sr_ore_t01": {
            "name": "The Instigator",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_autoheavy_vt_vr_ore_t01": {
            "name": "Vindertech Blazer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_neonglow_sr_ore_t01": {
            "name": "Neon Sniper Rifle",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_steampunk_drum_sr_ore_t01": {
            "name": "Duet",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_boombox_sr_ore_t01": {
            "name": "Surround Pound",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_semiauto_c_ore_t01": {
            "name": "Predator",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_blunt_heavy_flintlock_r_ore_t01": {
            "name": "Sir Hootie",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_hydraulic_sr_ore_t01": {
            "name": "Pulverizer",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_auto_uc_ore_t01": {
            "name": "Extinguisher",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_explosive_launcher_neonglow_sr_ore_t01": {
            "name": "Noble Launcher",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_sixshooter_c_ore_t01": {
            "name": "Nighthawk",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_retroscifi_vr_ore_t01": {
            "name": "Atomic Light Expander",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_medium_laser_founders_r_ore_t01": {
            "name": "Founder's Blazing Masamune",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_light_uc_ore_t01": {
            "name": "Ski Cleaver",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_pistol_winter_gingerbread_sr_ore_t01": {
            "name": "Ginger Blaster",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_semiauto_r_ore_t01": {
            "name": "Stampede",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_launcher_scavenger_sr_ore_t01": {
            "name": "Trash Cannon",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_ratrod_autodrum_sr_ore_t01": {
            "name": "Swan",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_handcannon_de_vr_ore_t01": {
            "name": "Bald Eagle",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_winter_sr_ore_t01": {
            "name": "Popshot",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_flintlock_vr_ore_t01": {
            "name": "Dragoon",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_auto_vt_sr_ore_t01": {
            "name": "Vindertech Pulsar",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_amr_r_ore_t01": {
            "name": "Zapotron",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_flintlock_sr_ore_t01": {
            "name": "Fishing Hook",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_medieval_vr_ore_t01": {
            "name": "Stalwart Squire",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_blackmetal_sr_ore_t01": {
            "name": "Husk Grinder",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_scythe_steampunk_vr_ore_t01": {
            "name": "Steam Thrasher",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_semiauto_founders_vr_ore_t01": {
            "name": "Founder's Vendetta",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_smg_unsilenced_military_r_ore_t01": {
            "name": "Specter",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_assault_singleshot_blackmetal_sr_ore_t01": {
            "name": "Blackout",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_military_auto_vr_ore_t01": {
            "name": "Bobcat",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_light_c_ore_t01": {
            "name": "Ski Cleaver",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_heavy_c_ore_t01": {
            "name": "Highlander",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_assault_singleshot_blackmetal_r_ore_t01": {
            "name": "Blackout",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_semiauto_sr_ore_t01": {
            "name": "Stampede",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_silenced_christmas_sr_ore_t01": {
            "name": "Cocoa .45",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_crossbow_military_sr_ore_t01": {
            "name": "Crossbow",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_scavenger_vr_ore_t01": {
            "name": "Pipe Down!",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_steampunk_bow_sr_ore_t01": {
            "name": "Cloudburst",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_blackmetal_vr_ore_t01": {
            "name": "Nightfall",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_military_burst_vr_ore_t01": {
            "name": "Lynx",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_artdeco_vr_ore_t01": {
            "name": "The Floor Flusher",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_light_bat_uc_ore_t01": {
            "name": "Mighty Slugger",
            "rarity": "uncommon",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_hydraulic_vr_ore_t01": {
            "name": "Pummeller",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_hydraulic_sr_ore_t01": {
            "name": "Pressure Cutter",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_rocket_winter_sr_ore_t01": {
            "name": "Santa's Little Helper",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_piercing_spear_scavenger_vr_ore_t01": {
            "name": "Head Clipper",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_sword_light_sr_ore_t01": {
            "name": "Stabsworth the III",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_blackmetal_vr_ore_t01": {
            "name": "Midnight Stalker",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_sniper_scavenger_vr_ore_t01": {
            "name": "Sunsetter",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_pistol_fleming_scoped_sr_ore_t01": {
            "name": "The Cyclops",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_explosive_medieval_sr_ore_t01": {
            "name": "The Jabberwocky",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_sniper_explosivebow_sr_ore_t01": {
            "name": "Boom Bow",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_firecracker_r_ore_t01": {
            "name": "Freedom's Herald",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_pistol_fleming_tactical_sr_ore_t01": {
            "name": "The Turncoat",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_artdeco_vr_ore_t01": {
            "name": "The Mouth Piece",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_vacuumtube_sr_ore_t01": {
            "name": "Vacuum Tube Rifle",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_break_sr_ore_t01": {
            "name": "Thunderbolt",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_standard_r_ore_t01": {
            "name": "Templar",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_blunt_tool_light_r_ore_t01": {
            "name": "Breaching Bar",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_heavy_c_ore_t01": {
            "name": "Crude Hammer",
            "rarity": "common",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_vacuumtube_sr_ore_t01": {
            "name": "Vacuum Tube Shotgun",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_auto_founders_vr_ore_t01": {
            "name": "Founder's Bearcat",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_flintlock_vr_ore_t01": {
            "name": "Walloper",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_fleming_sr_ore_t01": {
            "name": "The Grunt",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_blunt_hammer_rocket_vt_vr_ore_t01": {
            "name": "Vindertech Slammer",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_light_sr_ore_t01": {
            "name": "Claxe",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_surgical_drum_founders_r_ore_t01": {
            "name": "Founder's Drumroll",
            "rarity": "rare",
            "type": "weapon",
        },
        "Schematic:sid_sniper_medieval_sr_ore_t01": {
            "name": "Yeoman",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_edged_axe_medium_founders_vr_ore_t01": {
            "name": "Founder's Skullcleave",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_launcher_flintlock_sr_ore_t01": {
            "name": "Bowler",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_launcher_hydraulic_vr_ore_t01": {
            "name": "Dam Buster",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_assault_dragon_sr_ore_t01": {
            "name": "Dragon's Roar",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_assault_scavenger_sr_ore_t01": {
            "name": "Buzzcut",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Schematic:sid_pistol_autoheavy_vr_ore_t01": {
            "name": "Thrasher",
            "rarity": "epic",
            "type": "weapon",
        },
        "Schematic:sid_shotgun_tactical_precision_sr_ore_t01": {
            "name": "Nightclaw",
            "rarity": "legendary",
            "type": "weapon",
        },
        "Hero:hid_constructor_030_radconstructor_vr_t01": {
            "name": "Power Pop Penny",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_029_retroscifioutlander_vr_t01": {
            "name": "Cyberclops",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_sony_sr_t01": {
            "name": "Guardian Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_007hw_vr_t01": {
            "name": "Kyle the 13th",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_034_period_fleming_vr_t01": {
            "name": "Undercover Buzz",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_grenadegun_vr_t01": {
            "name": "Rescue Trooper Ramirez",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_rushbase_sr_t01": {
            "name": "B.A.S.E. Kyle",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_027_dinooutlander_sr_t01": {
            "name": "Fossil Southie",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_008_sr_t01": {
            "name": "Power B.A.S.E. Knox",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_036_dino_constructor_vr_t01": {
            "name": "Ankylo Kyle",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_040_pirate_br_vr_t01": {
            "name": "Buccaneer Ramirez",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_024_sr_t01": {
            "name": "Anti-Cuddle Sarah",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_030_radoutlander_sr_t01": {
            "name": "Main Stage Quinn",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_014_wukong_sr_t01": {
            "name": "Wukong",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_constructor_038_gumshoe_vr_t01": {
            "name": "Sleuth",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_xbox_sr_t01": {
            "name": "Guardian Knox",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_basehyper_vr_t01": {
            "name": "Hotfixer Hazard",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_xbox_r_t01": {
            "name": "Trailblazer Jess",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_014_m_sr_t01": {
            "name": "Berserker Renegade",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_zoneharvest_uc_t01": {
            "name": "Pathfinder Jess",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_commando_gunheadshothw_rs01_sr_t01": {
            "name": "Skull Ranger Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_009_m_sr_t01": {
            "name": "Centurion Hawk",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_010_f_vr_t01": {
            "name": "Deadly Blade Crash",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_021_sr_t01": {
            "name": "Ventura Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_basehyperhw_vr_t01": {
            "name": "Catstructor Penny",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_014_r_t01": {
            "name": "Riot Response Hazard",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_outlander_014_sr_t01": {
            "name": "Fireflower Eagle Eye",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_myth02_sr_t01": {
            "name": "Steel Wool Syd",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_constructor_040_lars_vr_t01": {
            "name": "Lars",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_gcgrenade_sr_t01": {
            "name": "Commando Spitfire",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_039_female_gumshoe_sr_t01": {
            "name": "Gumshoe",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_013_sr_t01": {
            "name": "Demolisher Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_016_m_v1_vr_t01": {
            "name": "Shockgunner Grizzly",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_037_director_sr_t01": {
            "name": "Director Riggs",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_009_vr_t01": {
            "name": "Vanguard Southie",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_010_sr_t01": {
            "name": "Deadly Star Scorpion",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_punchphase_r_t01": {
            "name": "Shock Specialist A.C.",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_myth03_sr_t01": {
            "name": "Carbide",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_ninja_016_m_v1_vr_t01": {
            "name": "Piercing Lotus Edge",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_017_f_v1_sr_t01": {
            "name": "Vintage-Tech Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_039_pajama_party_vr_t01": {
            "name": "Tricera Ops Ramirez",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_guntough_rs01_sr_t01": {
            "name": "Birthday Brigade Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_007_uc_t01": {
            "name": "Controller Harper",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_ninja_myth03_sr_t01": {
            "name": "Bladestorm Enforcer",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_constructor_024_pirateconstructor_sr_t01": {
            "name": "Privateer Hype",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_034_toymonkey_vr_t01": {
            "name": "Monks",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_013_vr_t01": {
            "name": "Warden Kyle",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_018_sr_t01": {
            "name": "Black Knight Garridan",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_outlander_007_sr_t01": {
            "name": "Enforcer Grizzly",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_guntough_valentine_sr_t01": {
            "name": "Love Ranger Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_035_f_cupid_sr_t01": {
            "name": "Stoneheart Farrah",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_starsrain_vr_t01": {
            "name": "Shuriken Master Sarah",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_010_m_sr_t01": {
            "name": "Trailblaster A.C.",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_punchdamage_vr_t01": {
            "name": "Striker A.C.",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_029_dinosoldier_sr_t01": {
            "name": "Rex Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_guntough_r_t01": {
            "name": "Survivalist Jonesy",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_outlander_015_f_v1_vr_t01": {
            "name": "T.E.D.D. Shot Jess",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_008_vr_t01": {
            "name": "Special Forces Banshee",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_guntough_vr_t01": {
            "name": "Survivalist Jonesy",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_038_male_ninja_pirate_vr_t01": {
            "name": "Freebooter Ken",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_015_f_v1_vr_t01": {
            "name": "Harvester Sarah",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_guntough_m_4thjuly_sr_t01": {
            "name": "Stars and Stripes Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_034_mechstructor_vr_t01": {
            "name": "Machinist Mina",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_zoneharvest_sr_t01": {
            "name": "Pathfinder Jess",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_020_sr_t01": {
            "name": "Cloaked Shadow",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_rushbase_r_t01": {
            "name": "B.A.S.E. Kyle",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_grenademaster_sr_t01": {
            "name": "Master Grenadier Ramirez",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_constructor_028_retroscificonstructor_vr_t01": {
            "name": "ED-EE",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_011_sr_t01": {
            "name": "Machinist Harper",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_038_clip_sr_t01": {
            "name": 'Cassie "Clip" Lipman',
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_gunheadshot_vr_t01": {
            "name": "Urban Assault Headhunter",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_zoneharvest_blockbuster_sr_t01": {
            "name": "Archaeolo-Jess",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_grenadegun_uc_t01": {
            "name": "Rescue Trooper Ramirez",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_commando_013_r_t01": {
            "name": "Demolisher Jonesy",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_plasmadamage_easter_sr_t01": {
            "name": "Miss Bunny Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_011_m_sr_t01": {
            "name": "Buckshot Raptor",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_025_pirateoutlander_vr_t01": {
            "name": "Crossbones Barret",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_024_sr_t01": {
            "name": "Steel Wool Anthony",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_outlander_010_sr_t01": {
            "name": "Shockblaster Buzz",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_023_sr_t01": {
            "name": "Jolly Headhunter",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_021_sr_t01": {
            "name": "Dark Vanguard Airheart",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_017_m_v1_vr_t01": {
            "name": "Wild Fragment Deadeye",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_042_assembler_sr_t01": {
            "name": "Robo-Ray",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_013_r_t01": {
            "name": "Warden Kyle",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_038_shock_wave_vr_t01": {
            "name": "Powerhouse",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_xbox_r_t01": {
            "name": "Guardian Knox",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_016_f_v1_vr_t01": {
            "name": "Double Agent Evelynn",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_042_major_vr_t01": {
            "name": "Major Oswald",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_017_m_v1_vr_t01": {
            "name": "Infiltrator Ken",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_040_assemblek_vr_t01": {
            "name": "Robo-Kevin",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_008_vr_t01": {
            "name": "Stonefoot Crash",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_013_sr_t01": {
            "name": "Alchemist Sarah",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_041_dinocollector_sr_t01": {
            "name": "Gia",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_044_fennix_sr_t01": {
            "name": "Fennix",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_slashbreath_vr_t01": {
            "name": "Dragon Scorch",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_xbox_sr_t01": {
            "name": "Trailblazer Jess",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_031_radsoldier_vr_t01": {
            "name": "Breakbeat Wildcat",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_sony_sr_t01": {
            "name": "Trailblazer A.C.",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_027_dinoconstructor_sr_t01": {
            "name": "Prehistoric Izza",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_011_r_t01": {
            "name": "Energy Thief Mari",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_037_junk_samurai_sr_t01": {
            "name": "Samurai Scrapper",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_013_stpatricks_f_v2_sr_t01": {
            "name": "Highland Warrior Wildcat",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_007_sr_t01": {
            "name": "Controller Harper",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_smokedimmak_sr_t01": {
            "name": "Dim Mak Mari",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_hammerplasma_vr_t01": {
            "name": "Electro-pulse Penny",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_slashtail_uc_t01": {
            "name": "Fleetfoot Ken",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_ninja_031_radninja_vr_t01": {
            "name": "Varsity Hiro",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_025_pirateninja_sr_t01": {
            "name": "Swashbuckler Keelhaul",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_009_sr_t01": {
            "name": "Colonel Wildcat",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_027_piratesoldier_vr_t01": {
            "name": "Buccaneer Jonesy",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_myth03_sr_t01": {
            "name": "Field Agent Rio",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_commando_007hw_rs01_sr_t01": {
            "name": "Brainiac Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_039_assemble_l_sr_t01": {
            "name": "Lok-Bot",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_punchphase_uc_t01": {
            "name": "Shock Specialist A.C.",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_commando_017_f_v1_vr_t01": {
            "name": "First Shot Rio",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_041_deco_sr_t01": {
            "name": "Venturion",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_xbox_sr_t01": {
            "name": "Jade Assassin Sarah",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_009_vr_t01": {
            "name": "Sentinel Hype",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_011_vr_t01": {
            "name": "Shrapnel Headhunter",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_014_sr_t01": {
            "name": "Riot Response Hazard",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_008_foundersm_sr_t01": {
            "name": "Special Forces Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_015_f_v1_r_t01": {
            "name": "T.E.D.D. Shot Jess",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_sony_vr_t01": {
            "name": "Commando Ramirez",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_028_razor_sr_t01": {
            "name": "Razor",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_009_sr_t01": {
            "name": "Skirmisher Edge",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_011_vr_t01": {
            "name": "Energy Thief Mari",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_013_sr_t01": {
            "name": "Fragment Flurry Jess",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_023_sr_t01": {
            "name": "Lynx Kassandra",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_commando_gcgrenade_t_vr_t01": {
            "name": "Sub Commando Jonesy",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_015_m_v1_sr_t01": {
            "name": "Bullet Storm Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_035_birthday_constructor_sr_t01": {
            "name": "Birthday Brigade Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_sony_r_t01": {
            "name": "Trailblazer A.C.",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_032_halloweensoldier_sr_t01": {
            "name": "JONESEE-BOT",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_044_gumshoe_dark_vr_t01": {
            "name": "Noir",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_039_dino_ninja_sr_t01": {
            "name": "Jurassic Ken",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_035_palespooky_outlander_holiday_vr_t01": {
            "name": "Red Willow",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_020_sr_t01": {
            "name": "Valkyrie Rio",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_zoneharvesthw_sr_t01": {
            "name": "Bloodfinder A.C.",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_032_toyconstructor_sr_t01": {
            "name": "Sgt. Tank Gatling",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_xbox_vr_t01": {
            "name": "Commando Renegade",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_017_f_v1_r_t01": {
            "name": "Vintage-Tech Penny",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_008_foundersf_sr_t01": {
            "name": "Special Forces Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_plasmadamage_sr_t01": {
            "name": "Plasma Specialist Izza",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_sony_sr_t01": {
            "name": "Bluestreak Ken",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_myth02_sr_t01": {
            "name": "The Cloaked Star",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_commando_011_r_t01": {
            "name": "Shrapnel Headhunter",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_starsrainhw_vr_t01": {
            "name": "Swift Shuriken Llamurai",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_033_halloweenquestsoldier_sr_t01": {
            "name": "Swamp Knight",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_sony_r_t01": {
            "name": "Commando Ramirez",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_033_halloweenninja_sr_t01": {
            "name": "Mermonster Ken",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_basehyper_sr_t01": {
            "name": "Hotfixer Hazard",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_041_pirate_02_br_vr_t01": {
            "name": "Sea Wolf Jonesy",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_zonepistol_vr_t01": {
            "name": "Ranger Deadeye",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_slashtail_sr_t01": {
            "name": "Fleetfoot Ken",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_015_f_v1_roadtrip_sr_t01": {
            "name": "Whiteout Fiona",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_015_f_v1_uc_t01": {
            "name": "Harvester Sarah",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_ninja_starsassassin_vr_t01": {
            "name": "Assassin Sarah",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_033_villain_fleming_vr_t01": {
            "name": "Gold Knox",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_013_stpatricks_m_sr_t01": {
            "name": "Battle Hound Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_myth02_sr_t01": {
            "name": "Raven",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_commando_grenadegun_sr_t01": {
            "name": "Rescue Trooper Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_017_m_v1_r_t01": {
            "name": "Infiltrator Ken",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_014_vr_t01": {
            "name": "Onslaught Headhunter",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_019_sr_t01": {
            "name": "Steel Wool Carlos",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_commando_guntough_uc_t01": {
            "name": "Survivalist Jonesy",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_outlander_008_vr_t01": {
            "name": "Recon Scout Eagle Eye",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_037_fuzzy_bear_teddy_sr_t01": {
            "name": "Metal Team Leader",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_036_dino_outlander_vr_t01": {
            "name": "Parasaur Jess",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_029_dinoninja_sr_t01": {
            "name": "Paleo Luna",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_026_sr_t01": {
            "name": "Fallen Love Ranger Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_010_sr_t01": {
            "name": "Guardian Bull",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_007_vr_t01": {
            "name": "Support Specialist Hawk",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_punchphase_sr_t01": {
            "name": "Shock Specialist A.C.",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_013_uc_t01": {
            "name": "Warden Kyle",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_constructor_011_r_t01": {
            "name": "Machinist Harper",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_shockdamage_r_t01": {
            "name": "Shock Trooper Renegade",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_030_retroscifisoldier_vr_t01": {
            "name": "Extraterrestrial Rio",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_guntough_f_4thjuly_sr_t01": {
            "name": "Star-Spangled Headhunter",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_011_m_easter_sr_t01": {
            "name": "Rabbit Raider Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_zonefragment_sr_t01": {
            "name": "Phase Scout Jess",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_commando_016_m_v1_sr_t01": {
            "name": "Undercover Vaughn",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_spherefragment_r_t01": {
            "name": "Trailblazer Quinn",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_038_gumshoe_sr_t01": {
            "name": "Sleuth",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_043_cranium_vr_t01": {
            "name": "Deimos",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_034_period_fleming_sr_t01": {
            "name": "Undercover Buzz",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_007hw_sr_t01": {
            "name": "Kyle the 13th",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_030_retroscifininja_vr_t01": {
            "name": "Intergalactic Ken",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_040_pirate_br_sr_t01": {
            "name": "Buccaneer Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_019_sr_t01": {
            "name": "Dire",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_ninja_014_vr_t01": {
            "name": "Whirlwind Scorch",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_023_sr_t01": {
            "name": "The Ice Queen",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_010_f_sr_t01": {
            "name": "Deadly Blade Crash",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_026_bunnyoutlander_sr_t01": {
            "name": "Cottontail Eagle Eye",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_036_dino_constructor_sr_t01": {
            "name": "Ankylo Kyle",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_gunheadshothw_vr_t01": {
            "name": "Skull Trooper Jonesy",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_029_retroscifioutlander_sr_t01": {
            "name": "Cyberclops",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_009_r_t01": {
            "name": "Skirmisher Edge",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_030_radconstructor_sr_t01": {
            "name": "Power Pop Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_sony_r_t01": {
            "name": "Bluestreak Ken",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_040_dennis_vr_t01": {
            "name": "Dennis",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_010_bday_sr_t01": {
            "name": "Birthday Brigade Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_zonepistolhw_vr_t01": {
            "name": "Beetlejess",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_010_m_4thjuly_sr_t01": {
            "name": "Old Glory A.C.",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_011_vr_t01": {
            "name": "Gunblazer Southie",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_rushbase_f_vr_t01": {
            "name": "Marathon Hype",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_034_toytinkerer_vr_t01": {
            "name": "Ted",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_026_bombsquadconstructor_vr_t01": {
            "name": "BombSquad Kyle",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_014_f_sr_t01": {
            "name": "Riot Control Izza",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_starsrainhw_sr_t01": {
            "name": "Swift Shuriken Llamurai",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_hammertank_vr_t01": {
            "name": "Tank Penny",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_starsrain_sr_t01": {
            "name": "Shuriken Master Sarah",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_swordmaster_sr_t01": {
            "name": "Swordmaster Ken",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_constructor_020_sr_t01": {
            "name": "Sentry Gunner Krampus",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_022_sr_t01": {
            "name": "Archetype Havoc",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_shockdamage_vr_t01": {
            "name": "Shock Trooper Renegade",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_009_sr_t01": {
            "name": "Vanguard Southie",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_031_halloweenconstructor_sr_t01": {
            "name": '"Arrlene" Izza',
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_015_vr_t01": {
            "name": "Heavy B.A.S.E. Kyle",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_smokedimmak_vr_t01": {
            "name": "Dim Mak Mari",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_039_pajama_party_sr_t01": {
            "name": "Tricera Ops Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_040_lars_sr_t01": {
            "name": "Lars",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_016_m_v1_sr_t01": {
            "name": "Shockgunner Grizzly",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_hammertank_r_t01": {
            "name": "Tank Penny",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_034_toymonkey_sr_t01": {
            "name": "Monks",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_punchdamage_sr_t01": {
            "name": "Striker A.C.",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_010_vr_t01": {
            "name": "Sergeant Jonesy",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_slashbreathhw_vr_t01": {
            "name": "Sarah Hotep",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_013_sr_t01": {
            "name": "Warden Kyle",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_013_stpatricks_f_v1_sr_t01": {
            "name": "Four Leaf Wildcat",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_017_f_v1_r_t01": {
            "name": "First Shot Rio",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_outlander_spherefragment_vr_t01": {
            "name": "Trailblazer Quinn",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_016_m_v1_sr_t01": {
            "name": "Piercing Lotus Edge",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_027_bunnyninja_sr_t01": {
            "name": "Dashing Hare Ken",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_024_vr_t01": {
            "name": "Sgt. Winter",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_016_f_v1_vr_t01": {
            "name": "Demolitionist Penny",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_011_uc_t01": {
            "name": "Shrapnel Headhunter",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_commando_015_m_v1_blockbuster_sr_t01": {
            "name": "Diecast Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_007_vr_t01": {
            "name": "Brawler Luna",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_007hw_vr_t01": {
            "name": "Ghoul Trooper Ramirez",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_034_mechstructor_sr_t01": {
            "name": "Machinist Mina",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_028_retroscificonstructor_sr_t01": {
            "name": "ED-EE",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_008_vr_t01": {
            "name": "Power B.A.S.E. Knox",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_017_m_v1_sr_t01": {
            "name": "Infiltrator Ken",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_009_m_r_t01": {
            "name": "Centurion Hawk",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_009_r_t01": {
            "name": "Colonel Wildcat",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_outlander_028_retroscifistoryoutlander_sr_t01": {
            "name": "Azalea Clark",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_outlander_009_r_t01": {
            "name": "Vanguard Southie",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_016_f_v1_sr_t01": {
            "name": "Double Agent Evelynn",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_015_f_v1_sr_t01": {
            "name": "T.E.D.D. Shot Jess",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_rushbase_vr_t01": {
            "name": "B.A.S.E. Kyle",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_xbox_vr_t01": {
            "name": "Guardian Knox",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_gunheadshot_sr_t01": {
            "name": "Urban Assault Headhunter",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_sony_vr_t01": {
            "name": "Guardian Penny",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_guntough_sr_t01": {
            "name": "Survivalist Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_017_m_v1_sr_t01": {
            "name": "Wild Fragment Deadeye",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_038_male_ninja_pirate_sr_t01": {
            "name": "Freebooter Ken",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_008_r_t01": {
            "name": "Stonefoot Crash",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_008_sr_t01": {
            "name": "Special Forces Banshee",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_014_m_sr_t01": {
            "name": "Flash A.C.",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_015_f_v1_sr_t01": {
            "name": "Harvester Sarah",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_myth02_sr_t01": {
            "name": "Ragnarok",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_ninja_022_sr_t01": {
            "name": "Overtaker Hiro",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_009_m_vr_t01": {
            "name": "Centurion Hawk",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_017_f_v1_vr_t01": {
            "name": "Vintage-Tech Penny",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_042_major_sr_t01": {
            "name": "Major Oswald",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_starsassassin_uc_t01": {
            "name": "Assassin Sarah",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_commando_007hw_sr_t01": {
            "name": "Ghoul Trooper Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_starsassassin_foundersm_sr_t01": {
            "name": "Lotus Assassin Ken",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_gcgrenade_vr_t01": {
            "name": "Commando Spitfire",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_009_r_t01": {
            "name": "Sentinel Hype",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_outlander_014_vr_t01": {
            "name": "Fireflower Eagle Eye",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_016_f_v1_sr_t01": {
            "name": "Deadly Lotus Luna",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_010_vr_t01": {
            "name": "Deadly Star Scorpion",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_038_shock_wave_sr_t01": {
            "name": "Powerhouse",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_030_radoutlander_vr_t01": {
            "name": "Main Stage Quinn",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_007_vr_t01": {
            "name": "Enforcer Grizzly",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_010_m_vr_t01": {
            "name": "Trailblaster A.C.",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_025_pirateoutlander_sr_t01": {
            "name": "Crossbones Barret",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_gunheadshot_starter_m_sr_t01": {
            "name": "Rogue Agent Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_016_f_v1_sr_t01": {
            "name": "Ambush Buzz",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_037_director_vr_t01": {
            "name": "Director Riggs",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_032_toyoutlander_sr_t01": {
            "name": "Jilly Teacup",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_039_female_gumshoe_vr_t01": {
            "name": "Gumshoe",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_024_pirateconstructor_vr_t01": {
            "name": "Privateer Hype",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_025_sr_t01": {
            "name": "Crackshot",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_slashbreath_sr_t01": {
            "name": "Dragon Scorch",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_040_assemblek_sr_t01": {
            "name": "Robo-Kevin",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_008_sr_t01": {
            "name": "Stonefoot Crash",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_023_sr_t01": {
            "name": "The Ice King",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_starsassassin_foundersf_sr_t01": {
            "name": "Lotus Assassin Sarah",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_slashtail_r_t01": {
            "name": "Fleetfoot Ken",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_hammerplasma_sr_t01": {
            "name": "Electro-pulse Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_011_sr_t01": {
            "name": "Shrapnel Headhunter",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_gcgrenade_r_t01": {
            "name": "Commando Spitfire",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_031_radsoldier_sr_t01": {
            "name": "Breakbeat Wildcat",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_017_f_v1_sr_t01": {
            "name": "First Shot Rio",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_008_r_t01": {
            "name": "Recon Scout Eagle Eye",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_outlander_038_clip_vr_t01": {
            "name": 'Cassie "Clip" Lipman',
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_031_radninja_sr_t01": {
            "name": "Varsity Hiro",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_011_vr_t01": {
            "name": "Machinist Harper",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_008_r_t01": {
            "name": "Special Forces Banshee",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_018_sr_t01": {
            "name": "Quickdraw Calamity",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_commando_007_uc_t01": {
            "name": "Support Specialist Hawk",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_commando_027_piratesoldier_sr_t01": {
            "name": "Buccaneer Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_007_r_t01": {
            "name": "Brawler Luna",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_013_vr_t01": {
            "name": "Demolisher Jonesy",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_013_stpatricks_sr_t01": {
            "name": "Staredown Southie",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_009_f_valentine_sr_t01": {
            "name": "Snuggle Specialist Sarah",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_029_radstoryconstructor_sr_t01": {
            "name": "Dennis Jr.",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_constructor_009_sr_t01": {
            "name": "Sentinel Hype",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_zonepistol_r_t01": {
            "name": "Ranger Deadeye",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_basehyperhw_sr_t01": {
            "name": "Catstructor Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_035_palespooky_outlander_holiday_sr_t01": {
            "name": "Red Willow",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_010_vr_t01": {
            "name": "Shockblaster Buzz",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_007_uc_t01": {
            "name": "Brawler Luna",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_commando_xbox_sr_t01": {
            "name": "Commando Renegade",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_018_sr_t01": {
            "name": "Plague Doctor Igor",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_hammertank_uc_t01": {
            "name": "Tank Penny",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_ninja_013_vr_t01": {
            "name": "Alchemist Sarah",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_042_assembler_vr_t01": {
            "name": "Robo-Ray",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_022_sr_t01": {
            "name": "Subzero Zenith",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_commando_sony_sr_t01": {
            "name": "Commando Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_gcgrenade_t_sr_t01": {
            "name": "Sub Commando Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_014_uc_t01": {
            "name": "Whirlwind Scorch",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_outlander_zoneharvest_vr_t01": {
            "name": "Pathfinder Jess",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_044_gumshoe_dark_sr_t01": {
            "name": "Noir",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_015_f_v1_r_t01": {
            "name": "Harvester Sarah",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_026_reddragonninja_sr_t01": {
            "name": "Hybrid",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_023_vr_t01": {
            "name": "Jolly Headhunter",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_016_m_v1_sr_t01": {
            "name": "Saboteur Bull",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_xbox_vr_t01": {
            "name": "Trailblazer Jess",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_011_sr_t01": {
            "name": "Energy Thief Mari",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_008_r_t01": {
            "name": "Power B.A.S.E. Knox",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_xbox_r_t01": {
            "name": "Commando Renegade",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_007_vr_t01": {
            "name": "Controller Harper",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_sony_r_t01": {
            "name": "Guardian Penny",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_011_f_v1_roadtrip_sr_t01": {
            "name": "Thunder Thora",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_025_progressivepirateconstructor_sr_t01": {
            "name": "Blakebeard The Blackhearted",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_constructor_039_assemble_l_vr_t01": {
            "name": "Lok-Bot",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_009_vr_t01": {
            "name": "Colonel Wildcat",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_011_f_v1_roadtrip_sr_t01": {
            "name": "Redline Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_019_sr_t01": {
            "name": "Sentry Gunner Airheart",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_041_deco_vr_t01": {
            "name": "Venturion",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_xbox_vr_t01": {
            "name": "Jade Assassin Sarah",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_zonepistol_sr_t01": {
            "name": "Ranger Deadeye",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_041_pirate_02_br_sr_t01": {
            "name": "Sea Wolf Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_014_r_t01": {
            "name": "Whirlwind Scorch",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_outlander_036_dino_outlander_sr_t01": {
            "name": "Parasaur Jess",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_starsassassin_sr_t01": {
            "name": "Assassin Sarah",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_grenadegun_r_t01": {
            "name": "Rescue Trooper Ramirez",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_025_pirateninja_vr_t01": {
            "name": "Swashbuckler Keelhaul",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_014_sr_t01": {
            "name": "Onslaught Headhunter",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_037_junk_samurai_vr_t01": {
            "name": "Samurai Scrapper",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_smokedimmak_r_t01": {
            "name": "Dim Mak Mari",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_rushbase_uc_t01": {
            "name": "B.A.S.E. Kyle",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_outlander_spherefragment_uc_t01": {
            "name": "Trailblazer Quinn",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_outlander_008_sr_t01": {
            "name": "Recon Scout Eagle Eye",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_033_villain_fleming_sr_t01": {
            "name": "Gold Knox",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_007_rs01_sr_t01": {
            "name": "Jingle Jess",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_014_vr_t01": {
            "name": "Riot Response Hazard",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_021_sr_t01": {
            "name": "Liteshow Spitfire",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_007_r_t01": {
            "name": "Enforcer Grizzly",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_028_bunnybrawler_sr_t01": {
            "name": "Bunny Brawler Luna",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_007_sr_t01": {
            "name": "Support Specialist Hawk",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_sony_vr_t01": {
            "name": "Trailblazer A.C.",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_plasmadamage_r_t01": {
            "name": "Plasma Specialist Izza",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_outlander_033_kurohomura_sr_t01": {
            "name": "Kurohomura",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_043_cranium_sr_t01": {
            "name": "Deimos",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_007_uc_t01": {
            "name": "Enforcer Grizzly",
            "rarity": "uncommon",
            "type": "hero",
        },
        "Hero:hid_outlander_018_sr_t01": {
            "name": "Sanguine Dusk",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_014_sr_t01": {
            "name": "Whirlwind Scorch",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_030_retroscifininja_sr_t01": {
            "name": "Intergalactic Ken",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_013_vr_t01": {
            "name": "Fragment Flurry Jess",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_035_birthday_constructor_vr_t01": {
            "name": "Birthday Brigade Penny",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_gunheadshothw_sr_t01": {
            "name": "Skull Trooper Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_basehyper_r_t01": {
            "name": "Hotfixer Hazard",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_grenadegun_m_t_sr_t01": {
            "name": "Rescue Trooper Havoc",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_015_m_v1_vr_t01": {
            "name": "Bullet Storm Jonesy",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_015_r_t01": {
            "name": "Heavy B.A.S.E. Kyle",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_xbox_r_t01": {
            "name": "Jade Assassin Sarah",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_007_r_t01": {
            "name": "Controller Harper",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_sony_vr_t01": {
            "name": "Bluestreak Ken",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_hammerplasma_4thjuly_sr_t01": {
            "name": "Patriot Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_031_halloweenoutlander_sr_t01": {
            "name": "Willow",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_basebig_sr_t01": {
            "name": "MEGA B.A.S.E. Kyle",
            "rarity": "mythic",
            "type": "hero",
        },
        "Hero:hid_outlander_zonepistolhw_sr_t01": {
            "name": "Beetlejess",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_plasmadamage_vr_t01": {
            "name": "Plasma Specialist Izza",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_039_dino_ninja_vr_t01": {
            "name": "Jurassic Ken",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_040_dennis_sr_t01": {
            "name": "Dennis",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_007_r_t01": {
            "name": "Support Specialist Hawk",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_021_sr_t01": {
            "name": "Forged Fate",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_032_toyconstructor_vr_t01": {
            "name": "Sgt. Tank Gatling",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_030_retroscifisoldier_sr_t01": {
            "name": "Extraterrestrial Rio",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_007_sr_t01": {
            "name": "Brawler Luna",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_010_sr_t01": {
            "name": "Sergeant Jonesy",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_punchphase_vr_t01": {
            "name": "Shock Specialist A.C.",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_013_r_t01": {
            "name": "Alchemist Sarah",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_commando_034_toytinkerer_sr_t01": {
            "name": "Ted",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_011_sr_t01": {
            "name": "Gunblazer Southie",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_008_foundersf_sr_t01": {
            "name": "Recon Scout Jess",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_rushbase_f_sr_t01": {
            "name": "Marathon Hype",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_008_foundersf_sr_t01": {
            "name": "Power B.A.S.E. Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_014_r_t01": {
            "name": "Fireflower Eagle Eye",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_slashbreathhw_sr_t01": {
            "name": "Sarah Hotep",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_037_fuzzy_bear_teddy_vr_t01": {
            "name": "Metal Team Leader",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_constructor_026_bombsquadconstructor_sr_t01": {
            "name": "BombSquad Kyle",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_045_chaos_agent_sr_t01": {
            "name": "Chaos Agent",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_slashtail_vr_t01": {
            "name": "Fleetfoot Ken",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_commando_shockdamage_sr_t01": {
            "name": "Shock Trooper Renegade",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_010_vr_t01": {
            "name": "Guardian Bull",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_outlander_spherefragment_sr_t01": {
            "name": "Trailblazer Quinn",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_009_vr_t01": {
            "name": "Skirmisher Edge",
            "rarity": "epic",
            "type": "hero",
        },
        "Hero:hid_ninja_starsassassin_r_t01": {
            "name": "Assassin Sarah",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_outlander_008_foundersm_sr_t01": {
            "name": "Recon Scout A.C.",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_008_foundersm_sr_t01": {
            "name": "Power B.A.S.E. Kyle",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_gunheadshot_m_v1_roadtrip_sr_t01": {
            "name": "Tactical Assault Sledgehammer",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_022_sr_t01": {
            "name": "Conqueror Magnus",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_024_sr_t01": {
            "name": "Sgt. Winter",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_015_sr_t01": {
            "name": "Heavy B.A.S.E. Kyle",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_015_f_v1_blockbuster_sr_t01": {
            "name": "Chromium Ramirez",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_commando_035_caped_valentine_sr_t01": {
            "name": "Dashing Hawk",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_ninja_014_f_sr_t01": {
            "name": "Thunderstrike Mari",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_outlander_zoneharvest_r_t01": {
            "name": "Pathfinder Jess",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_ninja_slashbreath_r_t01": {
            "name": "Dragon Scorch",
            "rarity": "rare",
            "type": "hero",
        },
        "Hero:hid_constructor_016_m_v1_blockbuster_sr_t01": {
            "name": "8-Bit Demo",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_016_f_v1_sr_t01": {
            "name": "Demolitionist Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Hero:hid_constructor_hammertank_sr_t01": {
            "name": "Tank Penny",
            "rarity": "legendary",
            "type": "hero",
        },
        "Worker:worker_halloween_husky_c_t01": {
            "name": "Husky",
            "rarity": "common",
            "type": "survivor",
        },
        "Worker:managerinventor_r_t01": {
            "name": "Lead Survivor",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:workerhalloween_uc_t01": {
            "name": "Husky",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:worker_halloween_pitcher_uc_t01": {
            "name": "Pitcher",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:managerengineer_c_t01": {
            "name": "Lead Survivor",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:managertrainer_uc_t01": {
            "name": "Lead Survivor",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managerexplorer_r_t01": {
            "name": "Lead Survivor",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managersoldier_sr_ramsie_t01": {
            "name": 'Joe "Ramsie" Bo',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:worker_joel_ur_t01": {
            "name": "Joel",
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:workerhalloween_alt_sr_t01": {
            "name": "Troll",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:managergadgeteer_sr_zapps_t01": {
            "name": 'Ellie "Zapps" Clerk',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:worker_halloween_husky_vr_t01": {
            "name": "Husky",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managermartialartist_uc_t01": {
            "name": "Lead Survivor",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managerinventor_sr_square_t01": {
            "name": 'Alfred "Square" Winestein',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managerdoctor_uc_t01": {
            "name": "Lead Survivor",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:worker_halloween_husk_c_t01": {
            "name": "Husk",
            "rarity": "common",
            "type": "survivor",
        },
        "Worker:managergadgeteer_sr_fixer_t01": {
            "name": 'Hamlet "Fixer" McKinney',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managerexplorer_vr_t01": {
            "name": "Lead Survivor",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:worker_halloween_pitcher_c_t01": {
            "name": "Pitcher",
            "rarity": "common",
            "type": "survivor",
        },
        "Worker:managergadgeteer_r_t01": {
            "name": "Lead Survivor",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managerexplorer_sr_eagle_t01": {
            "name": 'Amy "Eagle" Deerhart',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:workerbasic_sr_t01": {
            "name": "Survivor",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:workerbasic_c_t01": {
            "name": "Survivor",
            "rarity": "common",
            "type": "survivor",
        },
        "Worker:managerinventor_sr_rad_t01": {
            "name": 'Cary "Rad" Mary',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:worker_halloween_lobber_r_t01": {
            "name": "Lobber",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managertrainer_r_t01": {
            "name": "Lead Survivor",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managersoldier_c_t01": {
            "name": "Lead Survivor",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:workerhalloween_sr_t01": {
            "name": "Smasher",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:worker_halloween_lobber_vr_t01": {
            "name": "Lobber",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:worker_halloween_troll_sr_t01": {
            "name": "Troll",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:managertrainer_sr_raider_t01": {
            "name": 'Cara "Indiana" Loft',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managergadgeteer_sr_flak_t01": {
            "name": 'Kelly "Flak" Solek',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managerdoctor_sr_noctor_t01": {
            "name": 'Gregson "Noctor" Holmes',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managerinventor_vr_t01": {
            "name": "Lead Survivor",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:managerdoctor_r_t01": {
            "name": "Lead Survivor",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managerexplorer_uc_t01": {
            "name": "Lead Survivor",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managersoldier_sr_malcolm_t01": {
            "name": "Malcolm",
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managertrainer_sr_yoglattes_t01": {
            "name": 'Michelle "Yoglattes" Jillard',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managersoldier_vr_t01": {
            "name": "Lead Survivor",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:worker_halloween_husky_uc_t01": {
            "name": "Husky",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:managerengineer_vr_t01": {
            "name": "Lead Survivor",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:managerinventor_sr_frequency_t01": {
            "name": 'Nick "Frequency" Golub',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managertrainer_sr_jumpy_t01": {
            "name": 'Damon "Sprawl" Fitch',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managermartialartist_r_t01": {
            "name": "Lead Survivor",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managerinventor_c_t01": {
            "name": "Lead Survivor",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:workerhalloween_r_t01": {
            "name": "Pitcher",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managermartialartist_sr_samurai_t01": {
            "name": 'Mishiro "Samurai" Tofune',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:worker_halloween_husky_r_t01": {
            "name": "Husky",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managerexplorer_c_t01": {
            "name": "Lead Survivor",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:managerdoctor_sr_treky_t01": {
            "name": 'May "Treky" Jenson',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managerengineer_sr_countess_t01": {
            "name": 'Lacey "Countess" Ads',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managerengineer_r_t01": {
            "name": "Lead Survivor",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managermartialartist_sr_dragon_t01": {
            "name": 'Sie "Dragon" Lung',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managerexplorer_sr_birdie_t01": {
            "name": 'Saka "Birdie" Gale',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managergadgeteer_c_t01": {
            "name": "Lead Survivor",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:managerengineer_sr_maths_t01": {
            "name": 'Emma "Maths" Nolen',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managersoldier_uc_t01": {
            "name": "Lead Survivor",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managerengineer_uc_t01": {
            "name": "Lead Survivor",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managersoldier_sr_princess_t01": {
            "name": 'Lacy "Princess" Luw',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managerinventor_uc_t01": {
            "name": "Lead Survivor",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:workerbasic_vr_t01": {
            "name": "Survivor",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managergadgeteer_vr_t01": {
            "name": "Lead Survivor",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:worker_halloween_lobber_sr_t01": {
            "name": "Lobber",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:worker_halloween_pitcher_r_t01": {
            "name": "Pitcher",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managertrainer_c_t01": {
            "name": "Lead Survivor",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:worker_leprechaun_vr_t01": {
            "name": "Leprechaun",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managerengineer_sr_sobs_t01": {
            "name": "Jeev Sobs",
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:worker_halloween_husk_uc_t01": {
            "name": "Husk",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:workerbasic_r_t01": {
            "name": "Survivor",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:managerdoctor_sr_kingsly_t01": {
            "name": "Marlin Kingsly",
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managermartialartist_sr_tiger_t01": {
            "name": 'Misha "Tiger" Yu',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:worker_karolina_ur_t01": {
            "name": "Karolina",
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:managermartialartist_vr_t01": {
            "name": "Lead Survivor",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:managerquestdoctor_r_t01": {
            "name": "Dr. Blackwell",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managertrainer_vr_t01": {
            "name": "Lead Survivor",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:managersoldier_r_t01": {
            "name": "Lead Survivor",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managerdoctor_c_t01": {
            "name": "Lead Survivor",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:workerhalloween_vr_t01": {
            "name": "Lobber",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managerdoctor_vr_t01": {
            "name": "Lead Survivor",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:managerexplorer_sr_spacebound_t01": {
            "name": 'Ned "SpaceBound" Legstrong',
            "rarity": "mythic",
            "type": "survivor",
        },
        "Worker:workerhalloween_c_t01": {
            "name": "Husk",
            "rarity": "common",
            "type": "survivor",
        },
        "Worker:worker_halloween_troll_vr_t01": {
            "name": "Troll",
            "rarity": "epic",
            "type": "survivor",
        },
        "Worker:managermartialartist_c_t01": {
            "name": "Lead Survivor",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Worker:worker_halloween_smasher_sr_t01": {
            "name": "Smasher",
            "rarity": "legendary",
            "type": "survivor",
        },
        "Worker:managergadgeteer_uc_t01": {
            "name": "Lead Survivor",
            "rarity": "rare",
            "type": "survivor",
        },
        "Worker:workerbasic_uc_t01": {
            "name": "Survivor",
            "rarity": "uncommon",
            "type": "survivor",
        },
        "Defender:did_defenderpistol_basic_f_r_t01": {
            "name": "Defender",
            "rarity": "rare",
            "type": "defender",
        },
        "Defender:did_defendershotgun_basic_sr_t01": {
            "name": "Defender",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defenderassault_basic_sr_t01": {
            "name": "Defender",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defendermelee_basic_vr_t01": {
            "name": "Defender",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defenderassault_basic_r_t01": {
            "name": "Defender",
            "rarity": "rare",
            "type": "defender",
        },
        "Defender:did_defendersniper_basic_m_sr_t01": {
            "name": "Defender",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defendersniper_basic_sr_t01": {
            "name": "Defender",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defendershotgun_basic_f_sr_t01": {
            "name": "Defender",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defenderassault_founders_vr_t01": {
            "name": "Jill",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defenderassault_basic_m_sr_t01": {
            "name": "Defender",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defendershotgun_basic_r_t01": {
            "name": "Defender",
            "rarity": "rare",
            "type": "defender",
        },
        "Defender:did_defenderpistol_basic_r_t01": {
            "name": "Defender",
            "rarity": "rare",
            "type": "defender",
        },
        "Defender:did_defendermelee_basic_uc_t01": {
            "name": "Defender",
            "rarity": "uncommon",
            "type": "defender",
        },
        "Defender:did_defendermelee_basic_r_t01": {
            "name": "Defender",
            "rarity": "rare",
            "type": "defender",
        },
        "Defender:did_defenderpistol_basic_vr_t01": {
            "name": "Defender",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defenderpistol_basic_f_vr_t01": {
            "name": "Defender",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defendermelee_basic_sr_t01": {
            "name": "Defender",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defendershotgun_basic_f_r_t01": {
            "name": "Defender",
            "rarity": "rare",
            "type": "defender",
        },
        "Defender:did_defenderassault_basic_c_t01": {
            "name": "Defender",
            "rarity": "common",
            "type": "defender",
        },
        "Defender:did_defenderassault_basic_vr_t01": {
            "name": "Defender",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defendershotgun_basic_vr_t01": {
            "name": "Defender",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defendersniper_basic_r_t01": {
            "name": "Defender",
            "rarity": "rare",
            "type": "defender",
        },
        "Defender:did_defendersniper_basic_vr_t01": {
            "name": "Defender",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defendersniper_basic_m_vr_t01": {
            "name": "Defender",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defenderassault_basic_m_r_t01": {
            "name": "Defender",
            "rarity": "rare",
            "type": "defender",
        },
        "Defender:did_defenderpistol_basic_uc_t01": {
            "name": "Defender",
            "rarity": "uncommon",
            "type": "defender",
        },
        "Defender:did_defendershotgun_basic_c_t01": {
            "name": "Defender",
            "rarity": "common",
            "type": "defender",
        },
        "Defender:did_defenderpistol_basic_c_t01": {
            "name": "Defender",
            "rarity": "common",
            "type": "defender",
        },
        "Defender:did_defendermelee_basic_c_t01": {
            "name": "Defender",
            "rarity": "common",
            "type": "defender",
        },
        "Defender:did_defenderassault_basic_m_vr_t01": {
            "name": "Defender",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defendershotgun_basic_f_vr_t01": {
            "name": "Defender",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defendersniper_basic_m_r_t01": {
            "name": "Defender",
            "rarity": "rare",
            "type": "defender",
        },
        "Defender:did_defenderpistol_basic_f_sr_t01": {
            "name": "Defender",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defenderassault_basic_uc_t01": {
            "name": "Defender",
            "rarity": "uncommon",
            "type": "defender",
        },
        "Defender:did_defenderpistol_basic_sr_t01": {
            "name": "Defender",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defenderpistol_founders_vr_t01": {
            "name": "Idris",
            "rarity": "epic",
            "type": "defender",
        },
        "Defender:did_defendersniper_basic_c_t01": {
            "name": "Defender",
            "rarity": "common",
            "type": "defender",
        },
        "Defender:did_defendermelee_basic_f_sr_t01": {
            "name": "Val",
            "rarity": "legendary",
            "type": "defender",
        },
        "Defender:did_defendersniper_basic_uc_t01": {
            "name": "Defender",
            "rarity": "uncommon",
            "type": "defender",
        },
        "Defender:did_defendershotgun_basic_uc_t01": {
            "name": "Defender",
            "rarity": "uncommon",
            "type": "defender",
        },
        "AccountResource:currency_mtxswap": {
            "name": "VBucks",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:campaign_event_currency": {
            "name": "Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_summer": {
            "name": "Summer Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:reagent_c_t03": {
            "name": "Eye of the Storm",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_schematic_sr": {
            "name": "Legendary Schematic Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_hero_r": {
            "name": "Rare Hero Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_outlander_weapon_vr": {
            "name": "Founders Outlander Epic Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:heroxp": {
            "name": "Hero XP",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_weapon_r": {
            "name": "Rare Weapon Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_ninja_weapon_vr": {
            "name": "Founders Ninja Epic Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_weapons": {
            "name": "Weapon Designs",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_lunar": {
            "name": "Lunar Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:reagent_c_t04": {
            "name": "Storm Shard",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_trap_sr": {
            "name": "Legendary Trap Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:athenabattlestar": {
            "name": "Battle Stars",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_promotion_weapons": {
            "name": "Weapon Supercharger",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_roadtrip": {
            "name": "Road Trip Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_starterweapons_bundle": {
            "name": "Founders Starter Weapons Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_trap_r": {
            "name": "Rare Trap Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_outlander_weapon_sr": {
            "name": "Founders Outlander Legendary Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:schematicxp": {
            "name": "Schematic XP",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:reagent_promotion_heroes": {
            "name": "Hero Supercharger",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_alteration_generic": {
            "name": "RE-PERK!",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:reagent_promotion_traps": {
            "name": "Trap Supercharger",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_evolverarity_vr": {
            "name": "Epic Flux",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_ninja_weapon_sr": {
            "name": "Founders Ninja Legendary Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_soldier_weapon_vr": {
            "name": "Founders Soldier Epic Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:athenaseasonalxp_large": {
            "name": "Season XP",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_outlander_bundle": {
            "name": "Founders Outlander Bundle Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_cardpack_2021anniversary": {
            "name": "2021 Birthday Llama Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:compendiumxp": {
            "name": "Battle Royale Compendium XP",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_spring": {
            "name": "Spring Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:reagent_alteration_upgrade_vr": {
            "name": "Epic PERK-UP!",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_defender_vr": {
            "name": "Epic Defender Token",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_defender_r": {
            "name": "Rare Defender Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:phoenixxp_reward": {
            "name": "Venture XP",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_custom_firecracker_r": {
            "name": "Firecracker Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:athenaseasonalxp": {
            "name": "Season XP",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_constructor_weapon_vr": {
            "name": "Founders Constructor Epic Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_manager_r": {
            "name": "Rare Lead Survivor Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_basicpack": {
            "name": "Mini Reward Llama",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_manager_vr": {
            "name": "Epic Lead Survivor Token",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_constructor_bundle": {
            "name": "Founders Constructor Bundle Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_melee_vr": {
            "name": "Epic Melee Weapon Token",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:reagent_evolverarity_sr": {
            "name": "Legendary Flux",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_soldier_weapon_sr": {
            "name": "Founders Soldier Legendary Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_alteration_gameplay_generic": {
            "name": "Core RE-PERK!",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_defender_sr": {
            "name": "Legendary Defender Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_traps": {
            "name": "Trap Designs",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_snowballs": {
            "name": "Snowflake Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:reagent_alteration_upgrade_sr": {
            "name": "Legendary PERK-UP!",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_schematic_r": {
            "name": "Rare Schematic Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_constructor_weapon_sr": {
            "name": "Founders Constructor Legendary Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_alteration_ele_fire": {
            "name": "FIRE-UP!",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:feather_persistentresource": {
            "name": "Feathers",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:athenabattlepassxp": {
            "name": "Battle Pass XP",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_melee_r": {
            "name": "Rare Melee Weapon Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:currency_xrayllama": {
            "name": "X-Ray Tickets",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_scavenger": {
            "name": "Scavenger Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_ninja_bundle": {
            "name": "Founders Ninja Bundle Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_manager_sr": {
            "name": "Legendary Lead Survivor Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:specialcurrency_daily": {
            "name": "Daily Coins",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_stormzone": {
            "name": "Storm Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_hero_vr": {
            "name": "Epic Hero Token",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_melee_sr": {
            "name": "Legendary Melee Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_pumpkinwarhead": {
            "name": "Material - Pumpkin Warhead",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_weapon_vr": {
            "name": "Epic Weapon Token",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:phoenixxp": {
            "name": "Venture XP",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_worker_vr": {
            "name": "Epic Survivor Token",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_candy": {
            "name": "Candy",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_adventure": {
            "name": "Pirate Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:reagent_alteration_ele_nature": {
            "name": "AMP-UP!",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_founders_soldier_bundle": {
            "name": "Founders Soldier Bundle Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_alteration_ele_water": {
            "name": "FROST-UP!",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_ranged_vr": {
            "name": "Epic Ranged Weapon Token",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_ranged_r": {
            "name": "Rare Ranged Weapon Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_herobuyback": {
            "name": "Hero Recruitment Voucher",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_founders": {
            "name": "Founder's Coins",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_blockbuster": {
            "name": "Blockbuster Tickets",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:athenaseasonalxp_med": {
            "name": "Season XP",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_worker_sr": {
            "name": "Legendary Survivor Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_c_t01": {
            "name": "Pure Drop of Rain",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_schematic_vr": {
            "name": "Epic Schematic Token",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_weapon_sr": {
            "name": "Legendary Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_cardpack_bronze": {
            "name": "Upgrade Llama Token",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:reagent_people": {
            "name": "Training Manual",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:reagent_alteration_upgrade_r": {
            "name": "Rare PERK-UP!",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_hero_sr": {
            "name": "Legendary Hero Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_worker_r": {
            "name": "Rare Survivor Token",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:personnelxp": {
            "name": "Survivor XP",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:reagent_c_t02": {
            "name": "Lightning in a Bottle",
            "rarity": "epic",
            "type": "accountresource",
        },
        "AccountResource:athenawellrestedbooster": {
            "name": "Well Rested Booster",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:eventcurrency_scaling": {
            "name": "Gold",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_ranged_sr": {
            "name": "Legendary Ranged Weapon Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:reagent_promotion_survivors": {
            "name": "Survivor Supercharger",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:peopleresource": {
            "name": "People",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:athenalevelup": {
            "name": "Level Up",
            "rarity": "common",
            "type": "accountresource",
        },
        "AccountResource:reagent_alteration_upgrade_uc": {
            "name": "Uncommon PERK-UP!",
            "rarity": "uncommon",
            "type": "accountresource",
        },
        "AccountResource:reagent_evolverarity_r": {
            "name": "Rare Flux",
            "rarity": "rare",
            "type": "accountresource",
        },
        "AccountResource:voucher_item_buyback": {
            "name": "Weapon Research Voucher",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_cardpack_jackpot": {
            "name": "Legendary Troll Stash Llama Token",
            "rarity": "legendary",
            "type": "accountresource",
        },
        "AccountResource:voucher_generic_trap_vr": {
            "name": "Epic Trap Token",
            "rarity": "epic",
            "type": "accountresource",
        },
        "ConsumableAccountItem:smallxpboost_gift": {
            "name": "Teammate XP Boost",
            "rarity": "uncommon",
            "type": "consumableaccountitem",
        },
        "ConsumableAccountItem:smallxpboost": {
            "name": "XP Boost",
            "rarity": "uncommon",
            "type": "consumableaccountitem",
        },
        "CardPack:cardpack_dragon_hammer_sr": {
            "name": "Dragon's Fist",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_martialartist_sr": {
            "name": "Mythic Martial Artist",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season11_winter_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2019_schematic_sr": {
            "name": "Epic Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_worker_c": {
            "name": "Common Survivor",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_stormzone_base_hype": {
            "name": "Copper Hydra",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_manager_sr": {
            "name": "Mythic Lead Survivor",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_winter": {
            "name": "Winter Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_outlander_bundle": {
            "name": "Founders Outlander Bundle",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_wrapper_choice_test_basic": {
            "name": "Basic Choice Pack (Test)",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2019_heroes_ur": {
            "name": "Epic Item",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_inventor_r": {
            "name": "Epic Inventor",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_jackpot_winter2018_tickets": {
            "name": "Legendary Troll Stash Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_defender_sr": {
            "name": "Legendary Defender",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_launcher_halloween_sr": {
            "name": "Jack-O-Launcher",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defender_sr": {
            "name": "Legendary Defender",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_shotgun_military_sr": {
            "name": "Ground Pounder Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_spear_c": {
            "name": "Common Spear Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_scavenger_melee_vr": {
            "name": "Epic Scavenger Melee Weapon Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_r": {
            "name": "Rare Ranged Weapon Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_pistol_handcannon_sr": {
            "name": "Vindertech Blaster Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_vacuumtube_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_scythe_sr": {
            "name": "Legendary Scythe Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_uc": {
            "name": "Uncommon Hero",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_shotgun_c": {
            "name": "Common Shotgun Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_spring_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_500": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season11_ww_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_ceiling_sr": {
            "name": "Legendary Ceiling Trap Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_trainer_r": {
            "name": "Epic Personal Trainer",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_tornado": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season10_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_silver_collectionbook": {
            "name": "Upgrade Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_soldier_sr": {
            "name": "Mythic Marksman",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_weapons_collectionbook": {
            "name": "Weapon Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_flamegrill_sr": {
            "name": "Flame Grill Floor Trap",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_tiki": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season10_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_rare_melee": {
            "name": "Super Melee Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_assault_r": {
            "name": "Rare Assault Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_sr": {
            "name": "Mythic Lead Survivor",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_hero_commando_vr": {
            "name": "Epic Soldier",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_shotgun_vr": {
            "name": "Epic Shotgun Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_achievement_banner_lootlegend": {
            "name": "Challenge Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_perk_r": {
            "name": "Rare Perk Cache",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_all_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_vr": {
            "name": "Epic Ranged Weapon Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_horde_banner_2018_twinepeaks": {
            "name": "Road Trip Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_battlebreakers_ramirezhero": {
            "name": "Rescue Trooper Ramirez",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_06x": {
            "name": "Six Pack Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_scavenger_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_sword_c": {
            "name": "Common Sword Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2018_hero_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_military": {
            "name": "Military Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_rad_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_generators": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_achievement_banner_playswellwithothers": {
            "name": "Challenge Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season11_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_reward": {
            "name": "Upgrade Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_worker_r": {
            "name": "Rare Survivor",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_schematic_r": {
            "name": "Rare Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_spring_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendermelee_sr": {
            "name": "Legendary Bruiser",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_roadtrip2018_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_pistol_handcannon_semi_sr": {
            "name": "Vindertech Burster Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_horde_banner_2018_cannyvalley": {
            "name": "Road Trip Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_sniper_military_sr": {
            "name": "Obliterator Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_shotgun_r": {
            "name": "Rare Shotgun Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2018_storm_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_floor_vr": {
            "name": "Epic Floor Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_pistol_sr": {
            "name": "Legendary Pistol Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_dragon_assault_auto_sr": {
            "name": "Dragon's Roar Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_doctor_sr": {
            "name": "Mythic Doctor",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_silver_halloween": {
            "name": "Silver Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_ninja_vr": {
            "name": "Epic Ninja",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_pirate_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_inventor_sr": {
            "name": "Mythic Inventor",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_starter_sniper_boltaction_scope_r": {
            "name": "Copper Scoped Bullseye",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_sniper_sr": {
            "name": "Legendary Sniper Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_battlebreakers_kylehero": {
            "name": "BASE Kyle",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_season12_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season11_ww_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_fortnitemares_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_500": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_dragon_shotgun_standard_sr": {
            "name": "Dragon's Might Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_weapon_sr": {
            "name": "Legendary Weapon Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_assault_smg_silenced_military_sr": {
            "name": "Silenced Specter Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season11_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_stormzone_banners_plankerton": {
            "name": "Storm Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_personnel": {
            "name": "People Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_fortnitemares_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_season12": {
            "name": "Spy Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_jackpot_event": {
            "name": "Legendary Troll Stash Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_steampunk_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_blackmetal_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_sword_sr": {
            "name": "Legendary Sword Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_spear_r": {
            "name": "Rare Spear Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_club_sr": {
            "name": "Legendary Club Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_jackpot_super": {
            "name": "Legendary Troll Loot Truck... Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_commando_r": {
            "name": "Rare Soldier",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2017_halloween": {
            "name": "Fortnitemares Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_all_r": {
            "name": "Rare Item",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2017_winter_1": {
            "name": "Winter Llama #1/12",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_survivor_vr": {
            "name": "Epic Person",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_sniper_c": {
            "name": "Common Sniper Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_gadgeteer_vr": {
            "name": "Legendary Gadgeteer",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2021_schematic_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_holiday_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_military_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_sniper_standard_scope_vr": {
            "name": "Vindertech Sniper Rifle Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_weapon_r": {
            "name": "Rare Weapon Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_constructor_weapon_sr": {
            "name": "Copper Deconstructor",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendershotgun_uc": {
            "name": "Uncommon Shotgunner",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_assault_auto_sr": {
            "name": "Ramirez Assault Rifle",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_1100": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_worker_sr": {
            "name": "Legendary Survivor",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_shotgun_standard_vr": {
            "name": "Vindertech Scattergun Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_lunar_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_neon_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_hero_commando_sr": {
            "name": "Legendary Soldier",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_pistol_rapid_vr": {
            "name": "Copper Founder's Revolt",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_pods": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_10x": {
            "name": "All The Llamas",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season10_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_smg_vr": {
            "name": "Epic SMG Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_pirate_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_pistol_uc": {
            "name": "Uncommon Pistol Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_basic": {
            "name": "Mini Reward Llama",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_pirate_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season11_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_700": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_trainer_vr": {
            "name": "Legendary Personal Trainer",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendersniper_uc": {
            "name": "Uncommon Sniper",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_spring_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_constructor_bundle": {
            "name": "Founders Constructor Bundle",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_constructor_uc": {
            "name": "Uncommon Constructor",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_sniper_r": {
            "name": "Rare Sniper Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_dragon_sniper_scope_sr": {
            "name": "Dragon's Claw Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_halloweenworker_veryrare": {
            "name": "Epic Monster 'Cosplayer'",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season11_winter_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_tool_vr": {
            "name": "Epic Hardware Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_axe_uc": {
            "name": "Uncommon Axe Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_sniper_vr": {
            "name": "Epic Sniper Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_pistol_handcannon_de_military_sr": {
            "name": "Bald Eagle Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_assault_vr": {
            "name": "Epic Assault Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_gadgeteer_r": {
            "name": "Epic Gadgeteer",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter2018_banner_cooker": {
            "name": "Frostnite Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season12_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_c": {
            "name": "Common Melee Weapon Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_fortnitemares_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter2018_banner_iciclecluster": {
            "name": "Frostnite Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_sword_r": {
            "name": "Rare Sword Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_weapons_reward": {
            "name": "Weapon Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_club_uc": {
            "name": "Uncommon Club Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_commando_uc": {
            "name": "Uncommon Soldier",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_trapwalldarts_uc": {
            "name": "Wall Darts Trap Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter2018_banner_crownflame": {
            "name": "Frostnite Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_xp": {
            "name": "XP Candy Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2018_halloween": {
            "name": "Fortnitemares Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_700": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_200": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_schematic_stormking_ur": {
            "name": "Mythic Cache",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_schematic_uc": {
            "name": "Uncommon Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_spring_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_rad_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_overload": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_pirate_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_floor_sr": {
            "name": "Legendary Floor Trap Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_scavenger_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_rare_personnel": {
            "name": "Super People Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_explosive_launcher_military_sr": {
            "name": "Quad Launcher Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_jackpot_collectionbook": {
            "name": "Legendary Troll Stash Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_edged_sword_medium_vr": {
            "name": "Vindertech Sword Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_1000": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_pistol_autoheavy_vr": {
            "name": "Vindertech Repeater Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderassault_uc": {
            "name": "Uncommon Rifleman",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season9_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_edged_sword_neonglow_sr": {
            "name": "Krypton Sword Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_weapon_c": {
            "name": "Common Weapon Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_pistol_neonglow_sr": {
            "name": "Krypton Pistol Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_military_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_fortnitemares_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_gadgeteer_sr": {
            "name": "Mythic Gadgeteer",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_sniper_standard_scope_sr": {
            "name": "Vindertech Jolter Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season11_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_trap_wall_vr": {
            "name": "Epic Wall Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_holiday_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_ninja_sr": {
            "name": "Legendary Ninja",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendersniper_c": {
            "name": "Common Sniper",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_pirate_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_achievement_banner_unspeakablehorrors": {
            "name": "Challenge Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_sr": {
            "name": "Legendary Ranged Weapon Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_neon_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_season12_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_shotgun_sr": {
            "name": "Legendary Shotgun Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_blunt_hammer_rocket_vr": {
            "name": "Vindertech Rocket Hammer Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2019_heroes_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_meteor": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_axe_c": {
            "name": "Common Axe Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_pirate": {
            "name": "Pirate Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_wall_vr": {
            "name": "Epic Wall Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_uc": {
            "name": "Uncommon Trap Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_test2": {
            "name": "Rare Item (Test)",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_spring": {
            "name": "Lunar Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_mayday2019_banner_maydayiconnormal": {
            "name": "Hit the Road Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_crossbow_sr": {
            "name": "Heartbreaker",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_sword_vr": {
            "name": "Epic Sword Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_blackmetal_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_starter_edged_sword_medium_r": {
            "name": "Copper Shadow Blade",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_r": {
            "name": "Rare Hero",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_steampunk_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2018_winter": {
            "name": "Winter Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendersniper_r": {
            "name": "Rare Sniper",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_survivor_sr": {
            "name": "Legendary Person",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_tool_c": {
            "name": "Common Hardware Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_weapon_r": {
            "name": "Rare Weapon Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_manager_r": {
            "name": "Epic Lead Survivor",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_horde_banner_2018_stonewood": {
            "name": "Road Trip Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_basic_tutorial": {
            "name": "Mini Reward Llama",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_perk_vr": {
            "name": "Epic Perk Cache",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_pirate_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_materials_t01": {
            "name": "Evolution Materials Cache",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_sniper_crossbow_military_sr": {
            "name": "Crossbow Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_r": {
            "name": "Rare Melee Weapon Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_rare_ranged": {
            "name": "Super Ranged Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_shotgun_standard_sr": {
            "name": "Vindertech Disintegrator Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_spring_defendermelee_sr": {
            "name": "Bruiser Val",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_melee_r": {
            "name": "Rare Melee Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_200": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_weekly": {
            "name": "Weekly Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_explorer_vr": {
            "name": "Legendary Explorer",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderpistol_vr": {
            "name": "Epic Gunslinger",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_1300": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_lunar_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season12_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_outlander_vr": {
            "name": "Epic Outlander",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_artdeco_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_ceiling_r": {
            "name": "Rare Ceiling Trap Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season12_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_sniper_sr": {
            "name": "Legendary Sniper Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_03x": {
            "name": "Triple Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_founders": {
            "name": "Founder's Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_tool_sr": {
            "name": "Legendary Hardware Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_pistol_firecracker_sr": {
            "name": "Freedom's Herald Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_banners_t01": {
            "name": "Founders Banners: Standard",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_trainer_sr": {
            "name": "Mythic Personal Trainer",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_pirate_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_vindertech": {
            "name": "VIndertech Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_silver_xp": {
            "name": "Silver XP Candy Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_scythe_c": {
            "name": "Common Scythe Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_assault_sr": {
            "name": "Legendary Assault Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_achievement_banner_worldexplorer": {
            "name": "Challenge Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_spear_vr": {
            "name": "Epic Spear Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_stormzone_banners_stonewood": {
            "name": "Storm Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_stormzone_assault_hydra_sr": {
            "name": "Copper Hydra",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_achievement_banner_guardianangel": {
            "name": "Challenge Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season11_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze": {
            "name": "Upgrade Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderfoundersassault_vr": {
            "name": "Epic Founders Rifleman",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_toys_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_pistol_autoheavy_sr": {
            "name": "Vindertech Blazer Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_boombox_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_shadowops_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_spear_uc": {
            "name": "Uncommon Spear Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_shotgun_vr": {
            "name": "Epic Shotgun Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_achievement_banner_talentedbuilder": {
            "name": "Challenge Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_smg_sr": {
            "name": "Legendary SMG Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_vr": {
            "name": "Epic Hero",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_horde_banner_2018_plankerton": {
            "name": "Road Trip Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_materials_t02": {
            "name": "Evolution Materials Cache",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_trap_floor_vr": {
            "name": "Epic Floor Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_pistol_c": {
            "name": "Common Pistol Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_test3": {
            "name": "Epic Item (Test)",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_saucer": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_uc": {
            "name": "Uncommon Melee Weapon Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season11_winter_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_explorer_c": {
            "name": "Uncommon Explorer",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_hero_ninja_vr": {
            "name": "Epic Ninja",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_900": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_vr": {
            "name": "Epic Hero",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_gold": {
            "name": "Gold Cache",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_telebomb": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_holiday_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_wrapper_choice_test_multichoice": {
            "name": "MultiChoice Pack (Test)",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_perk_sr": {
            "name": "Legendary Perk Cache",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_assault_military_surgical_sr": {
            "name": "Tiger Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_blunt_hammer_rocket_sr": {
            "name": "Vindertech Slammer Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderpistol_c": {
            "name": "Common Gunslinger",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_tool_r": {
            "name": "Rare Hardware Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_tprifts": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_banners_t02": {
            "name": "Founders Banners: Deluxe",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_lava": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_uc": {
            "name": "Rare Lead Survivor",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_edged_sword_medium_sr": {
            "name": "Vindertech Slicer Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_sniper_neonglow_sr": {
            "name": "Neon Sniper Rifle Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defender_uc": {
            "name": "Uncommon Defender",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_wrapper_choice_test": {
            "name": "Choice Pack (Test)",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_materials_t03": {
            "name": "Evolution Materials Cache",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season9_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_900": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_pistol_r": {
            "name": "Rare Pistol Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_gumshoe_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_pirate_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_gold_ranged": {
            "name": "Gold Ranged Weapons Pack",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_military_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2019_heroes_sr": {
            "name": "Epic Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderpistol_sr": {
            "name": "Legendary Gunslinger",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendermelee_c": {
            "name": "Common Bruiser",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_axe_r": {
            "name": "Rare Axe Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_event_founders": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2018_roadtrip": {
            "name": "Road Trip Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_sword_sr": {
            "name": "Legendary Sword Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_personnel_reward": {
            "name": "People Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_axe_vr": {
            "name": "Epic Axe Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_schematic_r": {
            "name": "Rare Schematic Cache",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_explosive_launcher_neonglow_sr": {
            "name": "Noble Launcher Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_floor_freeze_vr": {
            "name": "Floor Freeze Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_400": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_halloween": {
            "name": "Fortnitemares Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_lunar": {
            "name": "Lunar Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_trap_wall_sr": {
            "name": "Legendary Wall Trap Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season9_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2020_artdeco_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter2018_banner_gingerbreadman": {
            "name": "Frostnite Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_cyberpunk_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_melee_vr": {
            "name": "Epic Melee Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter2018_banner_snowmanburied": {
            "name": "Frostnite Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_vindertech_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_winter_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season11_winter_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_1500": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_gold_xp": {
            "name": "Gold XP Candy Pack",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_04x": {
            "name": "Quadruple Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_sniper_uc": {
            "name": "Uncommon Sniper Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_wall_sr": {
            "name": "Legendary Wall Trap Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_scavenger": {
            "name": "Scavenger Choice Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_rad_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_400": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_assault_vr": {
            "name": "Epic Assault Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_banners_t03": {
            "name": "Founders Banners: Super Deluxe",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_constructor_r": {
            "name": "Rare Constructor",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendermelee_uc": {
            "name": "Uncommon Bruiser",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_jackpot": {
            "name": "Legendary Troll Stash Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season9": {
            "name": "Sci-Fi Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_tool_vr": {
            "name": "Epic Hardware Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_starter_assault_auto_r": {
            "name": "Copper Breacher",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season11_ww_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_melee": {
            "name": "Melee Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2021_heroes_ur": {
            "name": "Mythic Item",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_07x": {
            "name": "Lucky 7 Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendershotgun_vr": {
            "name": "Epic Shotgunner",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_trap_r": {
            "name": "Rare Trap Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_scythe_uc": {
            "name": "Uncommon Scythe Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendersniper_basic_m_vr": {
            "name": "Epic Sniper",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_sniper_standard_winter_sr": {
            "name": "Frostbite",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_gold_personnel": {
            "name": "Gold People Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_scythe_r": {
            "name": "Rare Scythe Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_artdeco_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_explorer_sr": {
            "name": "Mythic Explorer",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_smg_vr": {
            "name": "Epic SMG Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_silver_ranged": {
            "name": "Silver Ranged Weapons Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_halloweenworker_rare": {
            "name": "Rare Monster 'Cosplayer'",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_gold": {
            "name": "Gold Pack",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_materials_t04": {
            "name": "Evolution Materials Cache",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_outlander_sr": {
            "name": "Legendary Outlander",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_explorer_r": {
            "name": "Epic Explorer",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_explosive_r": {
            "name": "Rare Launcher Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_worker_uc": {
            "name": "Uncommon Survivor",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_shadowops_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_starter_floor_spikes_r": {
            "name": "Retractable Floor Spikes",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_martialartist_r": {
            "name": "Epic Martial Artist",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_holiday": {
            "name": "Holiday Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendermelee_r": {
            "name": "Rare Bruiser",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_constructor_vr": {
            "name": "Epic Constructor",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_weapon_uc": {
            "name": "Uncommon Weapon Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_shotgun_sr": {
            "name": "Legendary Shotgun Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_c": {
            "name": "Uncommon Lead Survivor",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_trap_floor_sr": {
            "name": "Legendary Floor Trap Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_hero_ninja_sr": {
            "name": "Legendary Ninja",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_boombox_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_sr": {
            "name": "Legendary Hero",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_ninja_r": {
            "name": "Rare Ninja",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_pistol_vr": {
            "name": "Epic Pistol Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter2018_banner_beartrapfoot": {
            "name": "Frostnite Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_gumshoe_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_axe_vr": {
            "name": "Epic Axe Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_hero_constructor_vr": {
            "name": "Epic Constructor",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_assault_neonglow_sr": {
            "name": "Argon Assault Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_silver_melee": {
            "name": "Silver Melee Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_spear_sr": {
            "name": "Legendary Spear Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season10_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_toys_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_lunar_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderpistol_r": {
            "name": "Rare Gunslinger",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendersniper_vr": {
            "name": "Epic Sniper",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season12_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_dragon_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_anniversary": {
            "name": "2nd Year Birthday Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_1000": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2017_winter": {
            "name": "Holiday Survival Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hydraulic_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_storm2018_banner_easy": {
            "name": "Into the Storm Survival Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_doctor_r": {
            "name": "Epic Doctor",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_sr": {
            "name": "Legendary Hero",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2020_artdeco_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_banners_t04": {
            "name": "Founders Banners: Limited",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_gold_melee": {
            "name": "Gold Melee Pack",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderassault_vr": {
            "name": "Epic Rifleman",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2020_artdeco": {
            "name": "Art Deco Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendershotgun_sr": {
            "name": "Legendary Shotgunner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_ceiling_falling_vr": {
            "name": "Ceiling Drop Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_edged_axe_neonglow_sr": {
            "name": "Argon Axe Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_military_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_scavenger_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderfounderspistol_vr": {
            "name": "Epic Founders Gunslinger",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_toys_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_engineer_vr": {
            "name": "Legendary Engineer",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_endurancewave30banner_theater01": {
            "name": "Stonewood Endurance Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defender_c": {
            "name": "Common Defender",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_axe_sr": {
            "name": "Legendary Axe Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_storm2018_banner_hard": {
            "name": "Into the Storm Survival Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter2018_banner_krampus": {
            "name": "Frostnite Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_weapon_vr": {
            "name": "Epic Weapon Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2021_anniversary": {
            "name": "2021 Birthday Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter_banner_easy": {
            "name": "Holiday Survival Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_hero_outlander_vr": {
            "name": "Epic Outlander",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_r": {
            "name": "Rare Hero",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_banners_t05": {
            "name": "Founders Banners: Ultimate",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderassault_c": {
            "name": "Common Rifleman",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_08x": {
            "name": "Octo-Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_schematic_vr": {
            "name": "Epic Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_commando_vr": {
            "name": "Epic Soldier",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_traps": {
            "name": "It's a Trap! Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_firststorm_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season10": {
            "name": "Rad Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_600": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season9_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_outlander_weapon_vr": {
            "name": "Copper Quickshot",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter2018_banner_wave30endgame": {
            "name": "Frostnite Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_cyberpunk_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2018_spring": {
            "name": "Spring Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_tool_sr": {
            "name": "Legendary Hardware Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_halloween_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_endurancewave30banner_theater02": {
            "name": "Plankerton Endurance Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_gold_traps": {
            "name": "Gold Trap Pack",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season9_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_dragon_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_100": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_assault_sr": {
            "name": "Legendary Assault Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_axe_sr": {
            "name": "Legendary Axe Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_rad_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_traps_collectionbook": {
            "name": "Trap Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_firststorm": {
            "name": "Storm Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_600": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_trap_vr": {
            "name": "Epic Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season11_ww_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season11_ww": {
            "name": "Wild West Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defender_r": {
            "name": "Rare Defender",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_r": {
            "name": "Epic Lead Survivor",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderassault_r": {
            "name": "Rare Rifleman",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_silver_traps": {
            "name": "Silver Trap Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_vr": {
            "name": "Epic Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2018_schematic_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_melee_sr": {
            "name": "Legendary Melee Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_shotgun_neonglow_sr": {
            "name": "Helium Shotgun Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_uc": {
            "name": "Uncommon Ranged Weapon Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter_banner_med": {
            "name": "Holiday Survival Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter_banner_hard": {
            "name": "Holiday Survival Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_shotgun_uc": {
            "name": "Uncommon Shotgun Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_club_vr": {
            "name": "Epic Club Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendershotgun_c": {
            "name": "Common Shotgunner",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_medieval_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_winter_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_defender_r": {
            "name": "Rare Defender",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_ninja_uc": {
            "name": "Uncommon Ninja",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_flintlock_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_club_c": {
            "name": "Common Club Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_ranged_vr": {
            "name": "Epic Ranged Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_vindertech_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_assault_lmg_military_sr": {
            "name": "HackSAW Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_100": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_smg_sr": {
            "name": "Legendary SMG Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_lunar_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_trap_ceiling_vr": {
            "name": "Epic Ceiling Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendersniper_sr": {
            "name": "Legendary Sniper",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_engineer_r": {
            "name": "Epic Engineer",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendershotgun_basic_f_vr": {
            "name": "Epic Shotgunner",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_survivor_uc": {
            "name": "Uncommon Person",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_halloween_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_roadtrip2018_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_05x": {
            "name": "Fiver Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_c": {
            "name": "Common Trap Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_hero_constructor_sr": {
            "name": "Legendary Constructor",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendershotgun_r": {
            "name": "Rare Shotgunner",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_club_r": {
            "name": "Rare Club Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_pistol_sr": {
            "name": "Legendary Pistol Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_gold_halloween": {
            "name": "Gold Pack",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season10_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_1200": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_spear_vr": {
            "name": "Epic Spear Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_soldier_bundle": {
            "name": "Founders Soldier Bundle",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_schematic_vr": {
            "name": "Epic Schematic Cache",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_vr": {
            "name": "Epic Melee Weapon Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_test_doublechoice1": {
            "name": "Double Choice Pack (Test 1)",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_scavenger_ranged_vr": {
            "name": "Epic Scavenger Ranged Weapon Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2018_anniversary": {
            "name": "1st Year Birthday Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_dragon_spring_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_scavenger_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_founders_nonexclusive": {
            "name": "Launch Event Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_weapon_sr": {
            "name": "Legendary Weapon Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_storm2018_banner_med": {
            "name": "Into the Storm Survival Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2021_heroes_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_constructor_sr": {
            "name": "Legendary Constructor",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_scythe_vr": {
            "name": "Epic Scythe Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_soldier_r": {
            "name": "Epic Marksman",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_ranged": {
            "name": "Ranged Weapon Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_silver_weapons": {
            "name": "Silver Weapons Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2018_storm": {
            "name": "Neon Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season11": {
            "name": "Fortnitemares Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hydraulic_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_edged_scythe_neonglow_sr": {
            "name": "Neon Scythe Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_edged_axe_vr": {
            "name": "Vindertech Axe Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_outlander_r": {
            "name": "Rare Outlander",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_soldier_weapon_vr": {
            "name": "Copper Drumroll",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_pistol_silenced_military_sr": {
            "name": "Whisper .45 Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_endurancewave30banner_theater03": {
            "name": "Canny Valley Endurance Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_starter_shotgun_standard_r": {
            "name": "Copper Crusader",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_ninja_weapon_vr": {
            "name": "Copper Blazing Masamune",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_assault_uc": {
            "name": "Uncommon Assault Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_outlander_weapon_sr": {
            "name": "Copper Quickshot",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_assault_lmg_neonglow_sr": {
            "name": "Mercury LMG Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_commando_sr": {
            "name": "Legendary Soldier",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_soldier_vr": {
            "name": "Legendary Marksman",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_survivor_c": {
            "name": "Common Person",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_firststorm_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_800": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_hero_outlander_sr": {
            "name": "Legendary Outlander",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_assault_auto_vr": {
            "name": "Vindertech Pulse Rifle Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_halloweenworker_uncommon": {
            "name": "Uncommon Monster 'Cosplayer'",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_starterweapons_bundle": {
            "name": "Copper Breacher",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_club_sr": {
            "name": "Legendary Club Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_wall_r": {
            "name": "Rare Wall Trap Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_schematic_sr": {
            "name": "Legendary Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_smg_uc": {
            "name": "Uncommon SMG Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_gold_weapons": {
            "name": "Gold Weapons Pack",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_vacuumtube_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_endurancewave30banner_theater04": {
            "name": "Twine Peaks Endurance Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_wrapper_unresolvedchoice": {
            "name": "Choice Card",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_season11_ww_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_r": {
            "name": "Rare Trap Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_engineer_sr": {
            "name": "Mythic Engineer",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2018_blockbuster": {
            "name": "Blockbuster Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_halloween_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2019_schematic_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_custom_firecracker_r": {
            "name": "Rare Firecracker Pistol",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_manager_vr": {
            "name": "Legendary Lead Survivor",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_weapons": {
            "name": "Weapon Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_defender_vr": {
            "name": "Epic Defender",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_800": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_toys_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_defender_vr": {
            "name": "Epic Defender",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_collectionbook": {
            "name": "Upgrade Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_vr": {
            "name": "Legendary Lead Survivor",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_rad": {
            "name": "Rad Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_test_doublechoice2": {
            "name": "Double Choice Pack (Test 2)",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season10_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_rare_heroes": {
            "name": "Super Hero Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_tool_uc": {
            "name": "Uncommon Hardware Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_floor_r": {
            "name": "Rare Floor Trap Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderassault_sr": {
            "name": "Legendary Rifleman",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_anomaly": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_smg_c": {
            "name": "Common SMG Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_ranged_r": {
            "name": "Rare Ranged Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_medieval_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_ranged_sr": {
            "name": "Legendary Ranged Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_achievement_banner_gognome": {
            "name": "Challenge Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_defendermelee_vr": {
            "name": "Epic Bruiser",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_scythe_vr": {
            "name": "Epic Scythe Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_trapceilinggas_uc": {
            "name": "Ceiling Poison Gas Trap Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2018_storm_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_pistol_handcannon_semi_vr": {
            "name": "Vindertech Burst Pistol Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2018_schematic_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_doctor_vr": {
            "name": "Legendary Doctor",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season11_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_basic_silver": {
            "name": "Mini Reward Llama",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_300": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_spear_sr": {
            "name": "Legendary Spear Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_roadtrip2018_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_sr": {
            "name": "Legendary Melee Weapon Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_sr": {
            "name": "Legendary Trap Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_ninja_bundle": {
            "name": "Founders Ninja Bundle",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_ceiling_vr": {
            "name": "Epic Ceiling Trap Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_test_doublechoice3": {
            "name": "Double Choice Pack (Test 3)",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_jackpot_superchoice": {
            "name": "Sm\u00f6rg\u00e5sbord Llama",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_silver": {
            "name": "Silver Pack",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_all_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_trap_sr": {
            "name": "Legendary Trap Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_survivor_r": {
            "name": "Rare Person",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_roadtrip2018_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_09x": {
            "name": "Niner Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_halloween_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_martialartist_vr": {
            "name": "Legendary Martial Artist",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_mtx_300": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_smg_r": {
            "name": "Rare SMG Schematic",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_defenderpistol_uc": {
            "name": "Uncommon Gunslinger",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_persistent_flintlock_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_2019_pirate": {
            "name": "Pirate Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_battlebreakers_jesshero": {
            "name": "Pathfinder Jess",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_sniper_vr": {
            "name": "Epic Sniper Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_silver_personnel": {
            "name": "Silver Survivor Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_winter2018_banner_tower": {
            "name": "Frostnite Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_season9_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_dragon_spring_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_manager_inventor_vr": {
            "name": "Legendary Inventor",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_schematic_c": {
            "name": "Common Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_hero_outlander_uc": {
            "name": "Uncommon Outlander",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_assault_c": {
            "name": "Common Assault Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_trap_walllauncher_fix": {
            "name": "Uncommon Trap Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_hero_persistent_holiday_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_scythe_sr": {
            "name": "Legendary Scythe Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_assault_auto_sr": {
            "name": "Vindertech Pulsar Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2018_hero_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_mayday2019_banner_maydayiconhard": {
            "name": "Hit the Road Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_wargames_banner_riotcity": {
            "name": "Wargames Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2021_schematic_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_schematic_sr": {
            "name": "Legendary Schematic Cache",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_halloweenworker_superrare": {
            "name": "Legendary Monster 'Cosplayer'",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary2021_heroes_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_ranged_pistol_vr": {
            "name": "Epic Pistol Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_sword_vr": {
            "name": "Epic Sword Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_stormzone_banners_cannyvalley": {
            "name": "Storm Event Banner",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_melee_club_vr": {
            "name": "Epic Club Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_melee_sword_uc": {
            "name": "Uncommon Sword Schematic",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_sniper_bbgun_sr": {
            "name": "Ralphie's Revenge",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_weapon_vr": {
            "name": "Epic Weapon Schematic",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_comp_eventscaling_1400": {
            "name": "Compensation Llama",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_bronze_02x": {
            "name": "Double Llama",
            "rarity": "rare",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_constructor_weapon_vr": {
            "name": "Copper Deconstructor",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_ninja_weapon_sr": {
            "name": "Copper Blazing Masamune",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_subchoice_trap_ceiling_sr": {
            "name": "Legendary Ceiling Trap Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_launcher_grenade_winter_sr": {
            "name": "Snowball Launcher",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_ranged_c": {
            "name": "Common Ranged Weapon Schematic",
            "rarity": "common",
            "type": "cardpack",
        },
        "CardPack:cardpack_vindertech_edged_axe_sr": {
            "name": "Vindertech Sever Schematic",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_event_persistent_fortnitemares": {
            "name": "Fortnitemares Llama",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_cache_perk_uc": {
            "name": "Uncommon Perk Cache",
            "rarity": "uncommon",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_schematics_season11_ww_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_founders_soldier_weapon_sr": {
            "name": "Copper Drumroll",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_worker_vr": {
            "name": "Epic Survivor",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary_heroes_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary_heroes_ur": {
            "name": "Mythic Item",
            "rarity": "mythic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary_heroes_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary_schematic_sr": {
            "name": "Legendary Item",
            "rarity": "legendary",
            "type": "cardpack",
        },
        "CardPack:cardpack_choice_anniversary_schematic_vr": {
            "name": "Epic Item",
            "rarity": "epic",
            "type": "cardpack",
        },
    },
    "ItemTypes": {
        "trap": "Trap",
        "weapon": "Weapon",
        "ammo": "Ammo",
        "ingredient": "Crafting Ingredient",
        "hero": "Hero",
        "defender": "Defender",
        "survivor": "Survivor",
        "accountresource": "Account Resource",
        "conditionalresource": "Account Resource",
        "consumableaccountitem": "Consumable",
        "cardpack": "Loot",
    },
    "SquadDetails": {
        "squad_attribute_medicine_emtsquad": {
            "name": "EMT Squad",
            "fort": "fortitude",
        },
        "squad_attribute_medicine_trainingteam": {
            "name": "Training Team",
            "fort": "fortitude",
        },
        "squad_attribute_arms_fireteamalpha": {
            "name": "Fire Team Alpha",
            "fort": "offense",
        },
        "squad_attribute_arms_closeassaultsquad": {
            "name": "Close Assault Squad",
            "fort": "offense",
        },
        "squad_attribute_scavenging_gadgeteers": {
            "name": "Gadgeteers",
            "fort": "resistance",
        },
        "squad_attribute_scavenging_scoutingparty": {
            "name": "Scouting Party",
            "fort": "resistance",
        },
        "squad_attribute_synthesis_corpsofengineering": {
            "name": "Corps Of Engineering",
            "fort": "tech",
        },
        "squad_attribute_synthesis_thethinktank": {
            "name": "The Think Tank",
            "fort": "tech",
        },
    },
    "LeadPreferredSquads": {
        "Homebase.Manager.IsDoctor": "squad_attribute_medicine_emtsquad",
        "Homebase.Manager.IsTrainer": "squad_attribute_medicine_trainingteam",
        "Homebase.Manager.IsSoldier": "squad_attribute_arms_fireteamalpha",
        "Homebase.Manager.IsMartialArtist": "squad_attribute_arms_closeassaultsquad",
        "Homebase.Manager.IsGadgeteer": "squad_attribute_scavenging_gadgeteers",
        "Homebase.Manager.IsExplorer": "squad_attribute_scavenging_scoutingparty",
        "Homebase.Manager.IsEngineer": "squad_attribute_synthesis_corpsofengineering",
        "Homebase.Manager.IsInventor": "squad_attribute_synthesis_thethinktank",
    },
    "LeadBonuses": {
        "Uncommon": (2, 0),
        "Rare": (3, 0),
        "Epic": (4, 0),
        "Legendary": (5, 0),
        "Mythic": (8, -2),
    },
    "SetBonuses": {
        "TrapDurability": {
            "name": "Trap Durability",
            "bonus": 8,
            "requirement": 2,
        },
        "RangedDamage": {
            "name": "Ranged Damage",
            "bonus": 5,
            "requirement": 3,
        },
        "MeleeDamage": {
            "name": "Melee Damage",
            "bonus": 5,
            "requirement": 3,
        },
        "TrapDamage": {
            "name": "Trap Damage",
            "bonus": 5,
            "requirement": 3,
        },
        "AbilityDamage": {
            "name": "Ability Damage",
            "bonus": 5,
            "requirement": 3,
        },
        "Fortitude": {
            "name": "Health",
            "bonus": 5,
            "requirement": 2,
        },
        "Resistance": {
            "name": "Shield",
            "bonus": 5,
            "requirement": 2,
        },
        "ShieldRegen": {
            "name": "Shield Regeneration",
            "bonus": 5,
            "requirement": 2,
        },
    },
    "ItemPowerLevels": {
        "Other": {
            "Common": {
                "0": {"1": 1},
                "1": {
                    "1": 1,
                    "2": 2,
                    "3": 3,
                    "4": 4,
                    "5": 5,
                    "6": 7,
                    "7": 8,
                    "8": 9,
                    "9": 10,
                    "10": 11,
                },
                "2": {
                    "10": 17,
                    "11": 18,
                    "12": 20,
                    "13": 21,
                    "14": 22,
                    "15": 23,
                    "16": 24,
                    "17": 26,
                    "18": 27,
                    "19": 28,
                    "20": 29,
                },
            },
            "Uncommon": {
                "1": {
                    "1": 4,
                    "2": 5,
                    "3": 6,
                    "4": 8,
                    "5": 9,
                    "6": 10,
                    "7": 11,
                    "8": 13,
                    "9": 14,
                    "10": 15,
                },
                "2": {
                    "10": 22,
                    "11": 23,
                    "12": 24,
                    "13": 26,
                    "14": 27,
                    "15": 28,
                    "16": 30,
                    "17": 31,
                    "18": 32,
                    "19": 34,
                    "20": 35,
                },
                "3": {
                    "20": 43,
                    "21": 45,
                    "22": 46,
                    "23": 47,
                    "24": 48,
                    "25": 50,
                    "26": 51,
                    "27": 53,
                    "28": 54,
                    "29": 55,
                    "30": 57,
                },
            },
            "Rare": {
                "1": {
                    "1": 7,
                    "2": 8,
                    "3": 9,
                    "4": 11,
                    "5": 12,
                    "6": 13,
                    "7": 15,
                    "8": 16,
                    "9": 17,
                    "10": 19,
                },
                "2": {
                    "10": 26,
                    "11": 27,
                    "12": 29,
                    "13": 30,
                    "14": 32,
                    "15": 33,
                    "16": 35,
                    "17": 36,
                    "18": 37,
                    "19": 39,
                    "20": 41,
                },
                "3": {
                    "20": 49,
                    "21": 51,
                    "22": 52,
                    "23": 54,
                    "24": 55,
                    "25": 56,
                    "26": 58,
                    "27": 59,
                    "28": 61,
                    "29": 62,
                    "30": 64,
                },
                "4": {
                    "30": 74,
                    "31": 75,
                    "32": 77,
                    "33": 78,
                    "34": 80,
                    "35": 81,
                    "36": 83,
                    "37": 84,
                    "38": 86,
                    "39": 87,
                    "40": 89,
                },
            },
            "Epic": {
                "1": {
                    "1": 9,
                    "2": 11,
                    "3": 12,
                    "4": 14,
                    "5": 15,
                    "6": 17,
                    "7": 18,
                    "8": 20,
                    "9": 21,
                    "10": 22,
                },
                "2": {
                    "10": 30,
                    "11": 32,
                    "12": 34,
                    "13": 35,
                    "14": 37,
                    "15": 39,
                    "16": 40,
                    "17": 42,
                    "18": 43,
                    "19": 45,
                    "20": 46,
                },
                "3": {
                    "20": 55,
                    "21": 56,
                    "22": 58,
                    "23": 60,
                    "24": 61,
                    "25": 63,
                    "26": 65,
                    "27": 66,
                    "28": 68,
                    "29": 70,
                    "30": 71,
                },
                "4": {
                    "30": 81,
                    "31": 83,
                    "32": 84,
                    "33": 86,
                    "34": 87,
                    "35": 89,
                    "36": 91,
                    "37": 92,
                    "38": 94,
                    "39": 96,
                    "40": 97,
                },
                "5": {
                    "40": 108,
                    "41": 110,
                    "42": 111,
                    "43": 112,
                    "44": 114,
                    "45": 115,
                    "46": 116,
                    "47": 118,
                    "48": 119,
                    "49": 120,
                    "50": 122,
                    "52": 124,
                    "54": 127,
                    "56": 130,
                    "58": 132,
                    "60": 135,
                },
            },
            "Legendary": {
                "1": {
                    "1": 12,
                    "2": 14,
                    "3": 15,
                    "4": 17,
                    "5": 18,
                    "6": 20,
                    "7": 21,
                    "8": 23,
                    "9": 24,
                    "10": 26,
                },
                "2": {
                    "10": 35,
                    "11": 36,
                    "12": 38,
                    "13": 40,
                    "14": 42,
                    "15": 43,
                    "16": 45,
                    "17": 46,
                    "18": 48,
                    "19": 50,
                    "20": 51,
                },
                "3": {
                    "20": 61,
                    "21": 62,
                    "22": 64,
                    "23": 66,
                    "24": 68,
                    "25": 69,
                    "26": 71,
                    "27": 73,
                    "28": 74,
                    "29": 76,
                    "30": 77,
                },
                "4": {
                    "30": 88,
                    "31": 90,
                    "32": 92,
                    "33": 94,
                    "34": 95,
                    "35": 97,
                    "36": 99,
                    "37": 101,
                    "38": 102,
                    "39": 104,
                    "40": 106,
                },
                "5": {
                    "40": 116,
                    "41": 117,
                    "42": 118,
                    "43": 120,
                    "44": 121,
                    "45": 123,
                    "46": 124,
                    "47": 126,
                    "48": 127,
                    "49": 129,
                    "50": 130,
                    "52": 133,
                    "54": 136,
                    "56": 139,
                    "58": 142,
                    "60": 144,
                },
            },
            "Mythic": {
                "1": {
                    "1": 12,
                    "2": 14,
                    "3": 15,
                    "4": 17,
                    "5": 18,
                    "6": 20,
                    "7": 21,
                    "8": 23,
                    "9": 24,
                    "10": 26,
                },
                "2": {
                    "10": 35,
                    "11": 36,
                    "12": 38,
                    "13": 40,
                    "14": 42,
                    "15": 43,
                    "16": 45,
                    "17": 46,
                    "18": 48,
                    "19": 50,
                    "20": 51,
                },
                "3": {
                    "20": 61,
                    "21": 62,
                    "22": 64,
                    "23": 66,
                    "24": 68,
                    "25": 69,
                    "26": 71,
                    "27": 73,
                    "28": 74,
                    "29": 76,
                    "30": 77,
                },
                "4": {
                    "30": 88,
                    "31": 90,
                    "32": 92,
                    "33": 94,
                    "34": 95,
                    "35": 97,
                    "36": 99,
                    "37": 101,
                    "38": 102,
                    "39": 104,
                    "40": 106,
                },
                "5": {
                    "40": 116,
                    "41": 117,
                    "42": 118,
                    "43": 120,
                    "44": 121,
                    "45": 123,
                    "46": 124,
                    "47": 126,
                    "48": 127,
                    "49": 129,
                    "50": 130,
                    "52": 133,
                    "54": 136,
                    "56": 139,
                    "58": 142,
                    "60": 144,
                },
            },
        },
        "Survivor": {
            "Common": {
                "1": {
                    "1": 1,
                    "2": 2,
                    "3": 3,
                    "4": 4,
                    "5": 5,
                    "6": 6,
                    "7": 7,
                    "8": 8,
                    "9": 9,
                    "10": 10,
                },
                "2": {
                    "10": 15,
                    "11": 16,
                    "12": 17,
                    "13": 18,
                    "14": 19,
                    "15": 20,
                    "16": 21,
                    "17": 22,
                    "18": 23,
                    "19": 24,
                    "20": 25,
                },
            },
            "Uncommon": {
                "1": {
                    "1": 5,
                    "2": 6,
                    "3": 7,
                    "4": 8,
                    "5": 9,
                    "6": 10,
                    "7": 11,
                    "8": 13,
                    "9": 14,
                    "10": 15,
                },
                "2": {
                    "10": 21,
                    "11": 22,
                    "12": 23,
                    "13": 24,
                    "14": 25,
                    "15": 26,
                    "16": 27,
                    "17": 29,
                    "18": 30,
                    "19": 31,
                    "20": 32,
                },
                "3": {
                    "20": 38,
                    "21": 39,
                    "22": 40,
                    "23": 41,
                    "24": 43,
                    "25": 44,
                    "26": 45,
                    "27": 46,
                    "28": 47,
                    "29": 48,
                    "30": 49,
                },
            },
            "Rare": {
                "1": {
                    "1": 10,
                    "2": 11,
                    "3": 12,
                    "4": 14,
                    "5": 15,
                    "6": 17,
                    "7": 18,
                    "8": 19,
                    "9": 20,
                    "10": 21,
                },
                "2": {
                    "10": 28,
                    "11": 29,
                    "12": 31,
                    "13": 32,
                    "14": 33,
                    "15": 34,
                    "16": 36,
                    "17": 37,
                    "18": 38,
                    "19": 39,
                    "20": 41,
                },
                "3": {
                    "20": 48,
                    "21": 49,
                    "22": 50,
                    "23": 51,
                    "24": 53,
                    "25": 54,
                    "26": 55,
                    "27": 56,
                    "28": 58,
                    "29": 59,
                    "30": 60,
                },
                "4": {
                    "30": 67,
                    "31": 68,
                    "32": 70,
                    "33": 71,
                    "34": 72,
                    "35": 73,
                    "36": 75,
                    "37": 76,
                    "38": 77,
                    "39": 78,
                    "40": 80,
                },
            },
            "Epic": {
                "1": {
                    "1": 15,
                    "2": 16,
                    "3": 18,
                    "4": 19,
                    "5": 20,
                    "6": 22,
                    "7": 23,
                    "8": 25,
                    "9": 26,
                    "10": 27,
                },
                "2": {
                    "10": 35,
                    "11": 37,
                    "12": 38,
                    "13": 39,
                    "14": 41,
                    "15": 42,
                    "16": 44,
                    "17": 45,
                    "18": 46,
                    "19": 48,
                    "20": 49,
                },
                "3": {
                    "20": 57,
                    "21": 58,
                    "22": 60,
                    "23": 61,
                    "24": 63,
                    "25": 64,
                    "26": 65,
                    "27": 67,
                    "28": 68,
                    "29": 69,
                    "30": 71,
                },
                "4": {
                    "30": 79,
                    "31": 80,
                    "32": 82,
                    "33": 83,
                    "34": 84,
                    "35": 86,
                    "36": 87,
                    "37": 88,
                    "38": 90,
                    "39": 91,
                    "40": 93,
                },
                "5": {
                    "40": 101,
                    "41": 102,
                    "42": 103,
                    "43": 105,
                    "44": 106,
                    "45": 107,
                    "46": 109,
                    "47": 110,
                    "48": 112,
                    "49": 113,
                    "50": 114,
                    "52": 117,
                    "54": 120,
                    "56": 123,
                    "58": 125,
                    "60": 128,
                },
            },
            "Legendary": {
                "1": {
                    "1": 20,
                    "2": 22,
                    "3": 23,
                    "4": 25,
                    "5": 26,
                    "6": 28,
                    "7": 29,
                    "8": 31,
                    "9": 32,
                    "10": 34,
                },
                "2": {
                    "10": 43,
                    "11": 44,
                    "12": 46,
                    "13": 47,
                    "14": 49,
                    "15": 50,
                    "16": 52,
                    "17": 53,
                    "18": 55,
                    "19": 56,
                    "20": 58,
                },
                "3": {
                    "20": 67,
                    "21": 68,
                    "22": 70,
                    "23": 71,
                    "24": 73,
                    "25": 74,
                    "26": 76,
                    "27": 77,
                    "28": 79,
                    "29": 80,
                    "30": 82,
                },
                "4": {
                    "30": 91,
                    "31": 92,
                    "32": 94,
                    "33": 95,
                    "34": 97,
                    "35": 98,
                    "36": 100,
                    "37": 101,
                    "38": 103,
                    "39": 104,
                    "40": 106,
                },
                "5": {
                    "40": 115,
                    "41": 116,
                    "42": 118,
                    "43": 119,
                    "44": 121,
                    "45": 122,
                    "46": 124,
                    "47": 125,
                    "48": 126,
                    "49": 128,
                    "50": 130,
                    "52": 133,
                    "54": 136,
                    "56": 139,
                    "58": 142,
                    "60": 145,
                },
            },
            "Mythic": {
                "1": {
                    "1": 25,
                    "2": 27,
                    "3": 28,
                    "4": 30,
                    "5": 32,
                    "6": 33,
                    "7": 35,
                    "8": 37,
                    "9": 38,
                    "10": 40,
                },
                "2": {
                    "10": 50,
                    "11": 51,
                    "12": 53,
                    "13": 55,
                    "14": 56,
                    "15": 58,
                    "16": 60,
                    "17": 61,
                    "18": 63,
                    "19": 64,
                    "20": 66,
                },
                "3": {
                    "20": 76,
                    "21": 78,
                    "22": 79,
                    "23": 81,
                    "24": 83,
                    "25": 84,
                    "26": 86,
                    "27": 87,
                    "28": 89,
                    "29": 91,
                    "30": 92,
                },
                "4": {
                    "30": 102,
                    "31": 104,
                    "32": 106,
                    "33": 107,
                    "34": 109,
                    "35": 110,
                    "36": 112,
                    "37": 114,
                    "38": 115,
                    "39": 117,
                    "40": 119,
                },
                "5": {
                    "40": 129,
                    "41": 130,
                    "42": 132,
                    "43": 133,
                    "44": 135,
                    "45": 137,
                    "46": 138,
                    "47": 140,
                    "48": 142,
                    "49": 143,
                    "50": 145,
                    "52": 148,
                    "54": 151,
                    "56": 154,
                    "58": 158,
                    "60": 161,
                },
            },
        },
        "LeadSurvivor": {
            "Uncommon": {
                "1": {
                    "1": 5,
                    "2": 6,
                    "3": 7,
                    "4": 8,
                    "5": 9,
                    "6": 10,
                    "7": 11,
                    "8": 12,
                    "9": 13,
                    "10": 14,
                },
                "2": {
                    "10": 19,
                    "11": 20,
                    "12": 21,
                    "13": 22,
                    "14": 23,
                    "15": 24,
                    "16": 25,
                    "17": 26,
                    "18": 27,
                    "19": 28,
                    "20": 29,
                },
                "3": {
                    "20": 34,
                    "21": 35,
                    "22": 36,
                    "23": 37,
                    "24": 38,
                    "25": 39,
                    "26": 40,
                    "27": 41,
                    "28": 42,
                    "29": 43,
                    "30": 44,
                },
            },
            "Rare": {
                "1": {
                    "1": 10,
                    "2": 11,
                    "3": 12,
                    "4": 13,
                    "5": 14,
                    "6": 15,
                    "7": 16,
                    "8": 18,
                    "9": 19,
                    "10": 20,
                },
                "2": {
                    "10": 26,
                    "11": 27,
                    "12": 28,
                    "13": 29,
                    "14": 30,
                    "15": 31,
                    "16": 33,
                    "17": 34,
                    "18": 35,
                    "19": 36,
                    "20": 37,
                },
                "3": {
                    "20": 43,
                    "21": 44,
                    "22": 45,
                    "23": 46,
                    "24": 48,
                    "25": 49,
                    "26": 50,
                    "27": 51,
                    "28": 52,
                    "29": 53,
                    "30": 54,
                },
                "4": {
                    "30": 60,
                    "31": 61,
                    "32": 63,
                    "33": 64,
                    "34": 65,
                    "35": 66,
                    "36": 67,
                    "37": 68,
                    "38": 69,
                    "39": 70,
                    "40": 72,
                },
            },
            "Epic": {
                "1": {
                    "1": 15,
                    "2": 16,
                    "3": 17,
                    "4": 19,
                    "5": 20,
                    "6": 21,
                    "7": 22,
                    "8": 24,
                    "9": 25,
                    "10": 26,
                },
                "2": {
                    "10": 33,
                    "11": 34,
                    "12": 36,
                    "13": 37,
                    "14": 38,
                    "15": 39,
                    "16": 41,
                    "17": 42,
                    "18": 43,
                    "19": 44,
                    "20": 46,
                },
                "3": {
                    "20": 53,
                    "21": 54,
                    "22": 55,
                    "23": 56,
                    "24": 58,
                    "25": 59,
                    "26": 60,
                    "27": 61,
                    "28": 63,
                    "29": 64,
                    "30": 65,
                },
                "4": {
                    "30": 72,
                    "31": 73,
                    "32": 75,
                    "33": 76,
                    "34": 77,
                    "35": 78,
                    "36": 80,
                    "37": 81,
                    "38": 82,
                    "39": 83,
                    "40": 85,
                },
                "5": {
                    "40": 92,
                    "41": 93,
                    "42": 94,
                    "43": 95,
                    "44": 97,
                    "45": 98,
                    "46": 99,
                    "47": 100,
                    "48": 102,
                    "49": 103,
                    "50": 104,
                    "52": 107,
                    "54": 110,
                    "56": 114,
                    "58": 117,
                    "60": 120,
                },
            },
            "Legendary": {
                "1": {
                    "1": 20,
                    "2": 21,
                    "3": 23,
                    "4": 24,
                    "5": 25,
                    "6": 27,
                    "7": 28,
                    "8": 30,
                    "9": 31,
                    "10": 32,
                },
                "2": {
                    "10": 40,
                    "11": 42,
                    "12": 43,
                    "13": 44,
                    "14": 46,
                    "15": 47,
                    "16": 49,
                    "17": 50,
                    "18": 51,
                    "19": 53,
                    "20": 54,
                },
                "3": {
                    "20": 62,
                    "21": 63,
                    "22": 65,
                    "23": 66,
                    "24": 68,
                    "25": 69,
                    "26": 70,
                    "27": 72,
                    "28": 73,
                    "29": 74,
                    "30": 76,
                },
                "4": {
                    "30": 84,
                    "31": 85,
                    "32": 87,
                    "33": 88,
                    "34": 89,
                    "35": 91,
                    "36": 92,
                    "37": 93,
                    "38": 95,
                    "39": 96,
                    "40": 98,
                },
                "5": {
                    "40": 106,
                    "41": 107,
                    "42": 108,
                    "43": 110,
                    "44": 111,
                    "45": 112,
                    "46": 114,
                    "47": 115,
                    "48": 117,
                    "49": 118,
                    "50": 119,
                    "52": 122,
                    "54": 125,
                    "56": 129,
                    "58": 132,
                    "60": 135,
                },
            },
            "Mythic": {
                "1": {
                    "1": 25,
                    "2": 27,
                    "3": 28,
                    "4": 30,
                    "5": 31,
                    "6": 33,
                    "7": 34,
                    "8": 36,
                    "9": 37,
                    "10": 39,
                },
                "2": {
                    "10": 48,
                    "11": 49,
                    "12": 51,
                    "13": 52,
                    "14": 54,
                    "15": 55,
                    "16": 57,
                    "17": 58,
                    "18": 60,
                    "19": 61,
                    "20": 63,
                },
                "3": {
                    "20": 72,
                    "21": 73,
                    "22": 75,
                    "23": 76,
                    "24": 78,
                    "25": 79,
                    "26": 81,
                    "27": 82,
                    "28": 84,
                    "29": 85,
                    "30": 87,
                },
                "4": {
                    "30": 96,
                    "31": 97,
                    "32": 99,
                    "33": 100,
                    "34": 102,
                    "35": 103,
                    "36": 105,
                    "37": 106,
                    "38": 108,
                    "39": 109,
                    "40": 111,
                },
                "5": {
                    "40": 120,
                    "41": 121,
                    "42": 123,
                    "43": 124,
                    "44": 126,
                    "45": 127,
                    "46": 129,
                    "47": 130,
                    "48": 132,
                    "49": 133,
                    "50": 135,
                    "52": 138,
                    "54": 141,
                    "56": 144,
                    "58": 147,
                    "60": 150,
                },
            },
        },
    },
}
